from __future__ import annotations

import inspect
import logging
import os
import pickle
import re
import sys
import warnings
from typing import Literal

import h5py
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
from torch_geometric.data.data import Data
from torch_geometric.data.dataset import Dataset
from tqdm import tqdm

from deeprank2.domain import edgestorage as Efeat
from deeprank2.domain import gridstorage
from deeprank2.domain import nodestorage as Nfeat
from deeprank2.domain import targetstorage as targets

_log = logging.getLogger(__name__)


class DeeprankDataset(Dataset):
    """Parent class of :class:`GridDataset` and :class:`GraphDataset`.

    This class inherits from :class:`torch_geometric.data.dataset.Dataset`.
    More detailed information about the parameters can be found in :class:`GridDataset` and :class:`GraphDataset`.
    """

    def __init__(
        self,
        hdf5_path: str | list[str],
        subset: list[str] | None,
        train_source: str | GridDataset | GraphDataset | None,
        target: str | None,
        target_transform: bool | None,
        target_filter: dict[str, str] | None,
        task: str | None,
        classes: list[str] | list[int] | list[float] | None,
        use_tqdm: bool,
        root: str,
        check_integrity: bool,
    ):
        super().__init__(root)

        if isinstance(hdf5_path, str):
            self.hdf5_paths = [hdf5_path]

        elif isinstance(hdf5_path, list):
            self.hdf5_paths = hdf5_path

        else:
            msg = f"hdf5_path: unexpected type: {type(hdf5_path)}"
            raise TypeError(msg)

        self.subset = subset
        self.train_source = train_source
        self.target = target
        self.target_transform = target_transform
        self.target_filter = target_filter

        if check_integrity:
            self._check_hdf5_files()

        self._check_task_and_classes(task, classes)
        self.use_tqdm = use_tqdm

        # create the indexing system
        # alows to associate each mol to an index
        # and get fname and mol name from the index
        self._create_index_entries()

        self.df = None
        self.means = None
        self.devs = None
        self.train_means = None
        self.train_devs = None

        # get the device
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    def _check_and_inherit_train(  # noqa: C901
        self,
        data_type: GridDataset | GraphDataset,
        inherited_params: list[str],
    ) -> None:
        """Check if the pre-trained model or training set provided are valid for validation and/or testing, and inherit the parameters."""
        if isinstance(self.train_source, str):
            try:
                if torch.cuda.is_available():
                    data = torch.load(self.train_source)
                else:
                    data = torch.load(self.train_source, map_location=torch.device("cpu"))
                if data["data_type"] is not data_type:
                    msg = (
                        f"The pre-trained model has been trained with data of type {data['data_type']}, but you are trying \n\t"
                        f"to define a {data_type}-class validation/testing dataset. Please provide a valid DeepRank2 \n\t"
                        f"model trained with {data_type}-class type data, or define the dataset using the appropriate class."
                    )
                    raise TypeError(msg)
                if data_type is GraphDataset:
                    self.train_means = data["means"]
                    self.train_devs = data["devs"]
                    # convert strings in 'transform' key to lambda functions
                    if data["features_transform"]:
                        for key in data["features_transform"].values():
                            if key["transform"] is None:
                                continue
                            key["transform"] = eval(key["transform"])  # noqa: S307, PGH001
            except pickle.UnpicklingError as e:
                msg = "The path provided to `train_source` is not a valid DeepRank2 pre-trained model."
                raise ValueError(msg) from e
        elif isinstance(self.train_source, data_type):
            data = self.train_source
            if data_type is GraphDataset:
                self.train_means = self.train_source.means
                self.train_devs = self.train_source.devs
        else:
            msg = (
                f"The train data provided is invalid: {type(self.train_source)}.\n\t"
                f"Please provide a valid training {data_type} or the path to a valid DeepRank2 pre-trained model."
            )
            raise TypeError(msg)

        # match parameters with the ones in the training set
        self._check_inherited_params(inherited_params, data)

    def _check_hdf5_files(self) -> None:
        """Checks if the data contained in the .HDF5 file is valid."""
        _log.info("\nChecking dataset Integrity...")
        to_be_removed = []
        for hdf5_path in self.hdf5_paths:
            try:
                with h5py.File(hdf5_path, "r") as f5:
                    entry_names = list(f5.keys())
                    if len(entry_names) == 0:
                        _log.info(f"    -> {hdf5_path} is empty ")
                        to_be_removed.append(hdf5_path)
            except Exception as e:  # noqa: BLE001, PERF203
                _log.error(e)
                _log.info(f"    -> {hdf5_path} is corrupted ")
                to_be_removed.append(hdf5_path)

        for hdf5_path in to_be_removed:
            self.hdf5_paths.remove(hdf5_path)

    def _check_task_and_classes(self, task: str, classes: str | None = None) -> None:
        if self.target in [targets.IRMSD, targets.LRMSD, targets.FNAT, targets.DOCKQ]:
            self.task = targets.REGRESS

        elif self.target in [targets.BINARY, targets.CAPRI]:
            self.task = targets.CLASSIF

        else:
            self.task = task

        if self.task not in [targets.CLASSIF, targets.REGRESS] and self.target is not None:
            msg = f"User target detected: {self.target} -> The task argument must be 'classif' or 'regress', currently set as {self.task}"
            raise ValueError(msg)

        if task != self.task and task is not None:
            warnings.warn(
                f"Target {self.target} expects {self.task}, but was set to task {task} by user.\nUser set task is ignored and {self.task} will be used.",
            )

        if self.task == targets.CLASSIF:
            if classes is None:
                self.classes = [0, 1]
                _log.info(f"Target classes set to: {self.classes}")
            else:
                self.classes = classes

            self.classes_to_index = {class_: index for index, class_ in enumerate(self.classes)}
        else:
            self.classes = None
            self.classes_to_index = None

    def _check_inherited_params(
        self,
        inherited_params: list[str],
        data: dict | GraphDataset | GridDataset,
    ) -> None:
        """Check if the parameters for validation and/or testing are the same as in the pre-trained model or training set provided.

        Args:
        inherited_params (List[str]): List of parameters that need to be checked for inheritance.
        data (Union[dict, class:`GraphDataset`, class:`GridDataset`]): The parameters in `inherited_param` will be inherited
        from the information contained in `data`.
        """
        self_vars = vars(self)
        if not isinstance(data, dict):
            data = vars(data)

        for param in inherited_params:
            if self_vars[param] != data[param]:
                if self_vars[param] != self.default_vars[param]:
                    _log.warning(
                        f"The {param} parameter set here is: {self_vars[param]}, "
                        f"which is not equivalent to the one in the training phase: {data[param]}./n"
                        f"Overwriting {param} parameter with the one used in the training phase.",
                    )
                setattr(self, param, data[param])

    def _create_index_entries(self) -> None:
        """Creates the indexing of each molecule in the dataset.

        Creates the indexing: [ ('1ak4.hdf5,1AK4_100w),...,('1fqj.hdf5,1FGJ_400w)].
        This allows to refer to one entry with its index in the list.
        """
        _log.debug(f"Processing data set with .HDF5 files: {self.hdf5_paths}")

        self.index_entries = []

        desc = f"   {self.hdf5_paths}{' dataset':25s}"
        if self.use_tqdm:
            hdf5_path_iterator = tqdm(self.hdf5_paths, desc=desc, file=sys.stdout)
        else:
            _log.info(f"   {self.hdf5_paths} dataset\n")
            hdf5_path_iterator = self.hdf5_paths
        sys.stdout.flush()

        for hdf5_path in hdf5_path_iterator:
            if self.use_tqdm:
                hdf5_path_iterator.set_postfix(entry_name=os.path.basename(hdf5_path))
            try:
                with h5py.File(hdf5_path, "r") as hdf5_file:
                    if self.subset is None:
                        entry_names = list(hdf5_file.keys())
                    else:
                        entry_names = [entry_name for entry_name in self.subset if entry_name in list(hdf5_file.keys())]

                    # skip self._filter_targets when target_filter is None, improve performance using list comprehension.
                    if self.target_filter is None:
                        self.index_entries += [(hdf5_path, entry_name) for entry_name in entry_names]
                    else:
                        self.index_entries += [(hdf5_path, entry_name) for entry_name in entry_names if self._filter_targets(hdf5_file[entry_name])]

            except Exception:  # noqa: BLE001
                _log.exception(f"on {hdf5_path}")

    def _filter_targets(self, grp: h5py.Group) -> bool:
        """Filters the entry according to a dictionary.

        The filter is based on the attribute self.target_filter that must be either
        of the form: { target_name : target_condition } or None.

        Args:
            grp (:class:`h5py.Group`): The entry group in the .HDF5 file.

        Returns:
            bool: True if we keep the entry False otherwise.

        Raises:
            ValueError: If an unsuported condition is provided.
        """
        if self.target_filter is None:
            return True

        for target_name, target_condition in self.target_filter.items():
            present_target_names = list(grp[targets.VALUES].keys())

            if target_name in present_target_names:
                # If we have a given target_condition, see if it's met.
                if isinstance(target_condition, str):
                    operation = target_condition
                    target_value = grp[targets.VALUES][target_name][()]
                    for operator_string in [">", "<", "==", "<=", ">=", "!="]:
                        operation = operation.replace(operator_string, f"{target_value}" + operator_string)

                    if not eval(operation):  # noqa: S307, PGH001
                        return False

                elif target_condition is not None:
                    msg = "Conditions not supported"
                    raise ValueError(msg, target_condition)

            else:
                _log.warning(f"   :Filter {target_name} not found for entry {grp}\n   :Filter options are: {present_target_names}")
        return True

    def len(self) -> int:
        """Gets the length of the dataset, either :class:`GridDataset` or :class:`GraphDataset` object.

        Returns:
            int: Number of complexes in the dataset.
        """
        return len(self.index_entries)

    def hdf5_to_pandas(  # noqa: C901
        self,
    ) -> pd.DataFrame:
        """Loads features data from the HDF5 files into a Pandas DataFrame in the attribute `df` of the class.

        Returns:
            :class:`pd.DataFrame`: Pandas DataFrame containing the selected features as columns per all data points in
                hdf5_path files.
        """
        df_final = pd.DataFrame()

        for fname in self.hdf5_paths:
            with h5py.File(fname, "r") as f:
                entry_name = next(iter(f.keys()))

                if self.subset is not None:
                    entry_names = [entry for entry, _ in f.items() if entry in self.subset]
                else:
                    entry_names = [entry for entry, _ in f.items()]

                df_dict = {}
                df_dict["id"] = entry_names

                for feat_type in self.features_dict:
                    for feat in self.features_dict[feat_type]:
                        # reset transform for each feature
                        transform = None
                        if self.features_transform:
                            transform = self.features_transform.get("all", {}).get("transform")
                            if (transform is None) and (feat in self.features_transform):
                                transform = self.features_transform.get(feat, {}).get("transform")
                        # Check the number of channels the features have
                        if f[entry_name][feat_type][feat][()].ndim == 2:  # noqa:PLR2004
                            for i in range(f[entry_name][feat_type][feat][:].shape[1]):
                                df_dict[feat + "_" + str(i)] = [f[entry_name][feat_type][feat][:][:, i] for entry_name in entry_names]
                                # apply transformation for each channel in this feature
                                if transform:
                                    df_dict[feat + "_" + str(i)] = [transform(row) for row in df_dict[feat + "_" + str(i)]]
                        else:
                            df_dict[feat] = [
                                f[entry_name][feat_type][feat][:] if f[entry_name][feat_type][feat][()].ndim == 1 else f[entry_name][feat_type][feat][()]
                                for entry_name in entry_names
                            ]
                            # apply transformation
                            if transform:
                                df_dict[feat] = [transform(row) for row in df_dict[feat]]

                df_temp = pd.DataFrame(data=df_dict)
            df_concat = pd.concat([df_final, df_temp])
        self.df = df_concat.reset_index(drop=True)
        return self.df

    def save_hist(  # noqa: C901
        self,
        features: str | list[str],
        fname: str = "features_hist.png",
        bins: int | list[float] | str = 10,
        figsize: tuple = (15, 15),
        log: bool = False,
    ) -> None:
        """After having generated a pd.DataFrame using hdf5_to_pandas method, histograms of the features can be saved in an image.

        Args:
            features (str | list[str]): Features to be plotted.
            fname (str): str or path-like or binary file-like object.
                Defaults to 'features_hist.png'.
            bins (int | list[float] | str, optional): If bins is an integer, it defines the number of equal-width bins in the range.
                If bins is a sequence, it defines the bin edges, including the left edge of the first bin and the right edge
                of the last bin; in this case, bins may be unequally spaced. All but the last (righthand-most) bin is half-open.
                If bins is a string, it is one of the binning strategies supported by numpy.histogram_bin_edges:
                'auto', 'fd', 'doane', 'scott', 'stone', 'rice', 'sturges', or 'sqrt'.
                Defaults to 10.
            figsize (tuple, optional): Saved figure sizes. Defaults to (15, 15).
            log (bool): Whether to apply log transformation to the data indicated by the `features` parameter. Defaults to False.
        """
        if self.df is None:
            self.hdf5_to_pandas()

        if not isinstance(features, list):
            features = [features]

        features_df = [col for feat in features for col in self.df.columns.to_numpy().tolist() if feat in col]

        means = [
            round(np.concatenate(self.df[feat].to_numpy()).mean(), 1)
            if isinstance(self.df[feat].to_numpy()[0], np.ndarray)
            else round(self.df[feat].to_numpy().mean(), 1)
            for feat in features_df
        ]
        devs = [
            round(np.concatenate(self.df[feat].to_numpy()).std(), 1)
            if isinstance(self.df[feat].to_numpy()[0], np.ndarray)
            else round(self.df[feat].to_numpy().std(), 1)
            for feat in features_df
        ]

        if len(features_df) > 1:
            fig, axs = plt.subplots(len(features_df), figsize=figsize)

            for row, feat in enumerate(features_df):
                if isinstance(self.df[feat].to_numpy()[0], np.ndarray):
                    if log:
                        log_data = np.log(np.concatenate(self.df[feat].to_numpy()))
                        log_data[log_data == -np.inf] = 0
                        axs[row].hist(log_data, bins=bins)
                    else:
                        axs[row].hist(np.concatenate(self.df[feat].to_numpy()), bins=bins)
                elif log:
                    log_data = np.log(self.df[feat].to_numpy())
                    log_data[log_data == -np.inf] = 0
                    axs[row].hist(log_data, bins=bins)
                else:
                    axs[row].hist(self.df[feat].to_numpy(), bins=bins)
                axs[row].set(
                    xlabel=f"{feat} (mean {means[row]}, std {devs[row]})",
                    ylabel="Count",
                )
            fig.tight_layout()

        elif len(features_df) == 1:
            fig = plt.figure(figsize=figsize)
            ax = fig.add_subplot(111)
            if isinstance(self.df[features_df[0]].to_numpy()[0], np.ndarray):
                if log:
                    log_data = np.log(np.concatenate(self.df[features_df[0]].to_numpy()))
                    log_data[log_data == -np.inf] = 0
                    ax.hist(log_data, bins=bins)
                else:
                    ax.hist(np.concatenate(self.df[features_df[0]].to_numpy()), bins=bins)
            elif log:
                log_data = np.log(self.df[features_df[0]].to_numpy())
                log_data[log_data == -np.inf] = 0
                ax.hist(log_data, bins=bins)
            else:
                ax.hist(self.df[features_df[0]].values, bins=bins)
            ax.set(
                xlabel=f"{features_df[0]} (mean {means[0]}, std {devs[0]})",
                ylabel="Count",
            )

        else:
            msg = "Please provide valid features names. They must be present in the current :class:`DeeprankDataset` children instance."
            raise ValueError(msg)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            fig.tight_layout()
            fig.savefig(fname)
            plt.close(fig)

    def _compute_mean_std(self) -> None:
        means = {
            col: round(np.nanmean(np.concatenate(self.df[col].values)), 1)
            if isinstance(self.df[col].to_numpy()[0], np.ndarray)
            else round(np.nanmean(self.df[col].to_numpy()), 1)
            for col in self.df.columns[1:]
        }
        devs = {
            col: round(np.nanstd(np.concatenate(self.df[col].to_numpy())), 1)
            if isinstance(self.df[col].to_numpy()[0], np.ndarray)
            else round(np.nanstd(self.df[col].to_numpy()), 1)
            for col in self.df.columns[1:]
        }
        self.means = means
        self.devs = devs


# Grid features are stored per dimension and named accordingly.
# Example: position_001, position_002, position_003 (for x,y,z)
# Use this regular expression to take the feature name apart
GRID_PARTIAL_FEATURE_NAME_PATTERN = re.compile(r"^([a-zA-Z_]+)_([0-9]{3})$")


class GridDataset(DeeprankDataset):
    """Class to load the .HDF5 files data into grids.

    Args:
        hdf5_path (str | list): Path to .HDF5 file(s). For multiple .HDF5 files, insert the paths in a list. Defaults to None.
        subset (list[str] | None, optional): list of keys from .HDF5 file to include. Defaults to None (meaning include all).
        train_source (str | class:`GridDataset` | None, optional): data to inherit information from the training dataset or the pre-trained model.
            If None, the current dataset is considered as the training set. Otherwise, `train_source` needs to be a dataset of the same class or
            the path of a DeepRank2 pre-trained model. If set, the parameters `features`, `target`, `traget_transform`, `task`, and `classes`
            will be inherited from `train_source`.
            Defaults to None.
        features (list[str] | str | Literal["all"] | None, optional): Consider all pre-computed features ("all") or some defined node features
            (provide a list, example: ["res_type", "polarity", "bsa"]). The complete list can be found in `deeprank2.domain.gridstorage`.
            Value will be ignored and inherited from `train_source` if `train_source` is assigned.
            Defaults to "all".
        target (str | None, optional): Default options are irmsd, lrmsd, fnat, binary, capri_class, and dockq. It can also be
            a custom-defined target given to the Query class as input (see: `deeprank2.query`); in this case,
            the task parameter needs to be explicitly specified as well.
            Only numerical target variables are supported, not categorical.
            If the latter is your case, please convert the categorical classes into
            numerical class indices before defining the :class:`GraphDataset` instance.
            Value will be ignored and inherited from `train_source` if `train_source` is assigned.
            Defaults to None.
        target_transform (bool, optional): Apply a log and then a sigmoid transformation to the target (for regression only).
            This puts the target value between 0 and 1, and can result in a more uniform target distribution and speed up the optimization.
            Value will be ignored and inherited from `train_source` if `train_source` is assigned.
            Defaults to False.
        target_filter (dict[str, str] | None, optional): Dictionary of type [target: cond] to filter the molecules.
            Note that the you can filter on a different target than the one selected as the dataset target.
            Defaults to None.
        task (Literal["regress", "classif"] | None, optional): 'regress' for regression or 'classif' for classification. Required if target not in
            ['irmsd', 'lrmsd', 'fnat', 'binary', 'capri_class', or 'dockq'], otherwise this setting is ignored.
            Automatically set to 'classif' if the target is 'binary' or 'capri_classes'.
            Automatically set to 'regress' if the target is 'irmsd', 'lrmsd', 'fnat', or 'dockq'.
            Value will be ignored and inherited from `train_source` if `train_source` is assigned.
            Defaults to None.
        classes (list[str] | list[int] | list[float] | None): Define the dataset target classes in classification mode.
            Value will be ignored and inherited from `train_source` if `train_source` is assigned.
            Defaults to None.
        use_tqdm (bool, optional): Show progress bar.
            Defaults to True.
        root (str, optional): Root directory where the dataset should be saved.
            Defaults to "./".
        check_integrity (bool, optional): Whether to check the integrity of the hdf5 files.
            Defaults to True.
    """

    def __init__(
        self,
        hdf5_path: str | list,
        subset: list[str] | None = None,
        train_source: str | GridDataset | None = None,
        features: list[str] | str | None = "all",
        target: str | None = None,
        target_transform: bool = False,
        target_filter: dict[str, str] | None = None,
        task: Literal["regress", "classif"] | None = None,
        classes: list[str] | list[int] | list[float] | None = None,
        use_tqdm: bool = True,
        root: str = "./",
        check_integrity: bool = True,
    ):
        super().__init__(
            hdf5_path,
            subset,
            train_source,
            target,
            target_transform,
            target_filter,
            task,
            classes,
            use_tqdm,
            root,
            check_integrity,
        )
        self.default_vars = {k: v.default for k, v in inspect.signature(self.__init__).parameters.items() if v.default is not inspect.Parameter.empty}
        self.default_vars["classes_to_index"] = None
        self.features = features
        self.target_transform = target_transform

        if train_source is not None:
            self.inherited_params = [
                "features",
                "target",
                "target_transform",
                "task",
                "classes",
                "classes_to_index",
            ]
            self._check_and_inherit_train(GridDataset, self.inherited_params)
            self._check_features()

        else:
            self._check_features()
            self.inherited_params = None

            try:
                fname, mol = self.index_entries[0]
            except IndexError as e:
                msg = "No entries found in the dataset. Please check the dataset parameters."
                raise IndexError(msg) from e
            with h5py.File(fname, "r") as f5:
                grp = f5[mol]
                possible_targets = grp[targets.VALUES].keys()
                if self.target is None:
                    msg = f"Please set the target during training dataset definition; targets present in the file/s are {possible_targets}."
                    raise ValueError(msg)
                if self.target not in possible_targets:
                    msg = f"Target {self.target} not present in the file/s; targets present in the file/s are {possible_targets}."
                    raise ValueError(msg)

        self.features_dict = {}
        self.features_dict[gridstorage.MAPPED_FEATURES] = self.features
        if self.target is not None:
            if isinstance(self.target, str):
                self.features_dict[targets.VALUES] = [self.target]
            else:
                self.features_dict[targets.VALUES] = self.target

    def _check_features(self) -> None:  # noqa: C901
        """Checks if the required features exist."""
        hdf5_path = self.hdf5_paths[0]

        # read available features
        with h5py.File(hdf5_path, "r") as f:
            mol_key = next(iter(f.keys()))
            if isinstance(self.features, list):
                self.features = [
                    GRID_PARTIAL_FEATURE_NAME_PATTERN.match(feature_name).group(1)
                    if GRID_PARTIAL_FEATURE_NAME_PATTERN.match(feature_name) is not None
                    else feature_name
                    for feature_name in self.features
                ]  # remove the dimension number suffix
                self.features = list(set(self.features))  # remove duplicates
            available_features = list(f[f"{mol_key}/{gridstorage.MAPPED_FEATURES}"].keys())
            available_features = [key for key in available_features if key[0] != "_"]  # ignore metafeatures

            hdf5_matching_feature_names = []  # feature names that match with the requested list of names
            unpartial_feature_names = []  # feature names without their dimension number suffix
            for feature_name in available_features:
                partial_feature_match = GRID_PARTIAL_FEATURE_NAME_PATTERN.match(feature_name)
                if partial_feature_match is not None:  # there's a dimension number in the feature name
                    unpartial_feature_name = partial_feature_match.group(1)

                    if self.features == "all" or (isinstance(self.features, list) and unpartial_feature_name in self.features):
                        hdf5_matching_feature_names.append(feature_name)

                    unpartial_feature_names.append(unpartial_feature_name)

                else:  # no numbers, it's a one-dimensional feature name
                    if self.features == "all" or (isinstance(self.features, list) and feature_name in self.features):
                        hdf5_matching_feature_names.append(feature_name)

                    unpartial_feature_names.append(feature_name)

        # check for the requested features
        missing_features = []
        if self.features == "all":
            self.features = sorted(available_features)
            self.default_vars["features"] = self.features
        else:
            if not isinstance(self.features, list):
                if self.features is None:
                    self.features = []
                else:
                    self.features = [self.features]
            for feature_name in self.features:
                if feature_name not in unpartial_feature_names:
                    _log.info(f"The feature {feature_name} was not found in the file {hdf5_path}.")
                    missing_features.append(feature_name)

            self.features = sorted(hdf5_matching_feature_names)

        # raise error if any features are missing
        if len(missing_features) > 0:
            msg = (
                f"Not all features could be found in the file {hdf5_path} under entry {mol_key}.\n\t"
                f"Missing features are: {missing_features}.\n\t"
                "Check feature_modules passed to the preprocess function.\n\t"
                "Probably, the feature wasn't generated during the preprocessing step.\n\t"
                f"Available features: {available_features}"
            )
            raise ValueError(msg)

    def get(self, idx: int) -> Data:
        """Gets one grid item from its unique index.

        Args:
            idx (int): Index of the item, ranging from 0 to len(dataset).

        Returns:
            :class:`torch_geometric.data.data.Data`: item with tensors x, y if present, entry_names.
        """
        file_path, entry_name = self.index_entries[idx]
        return self.load_one_grid(file_path, entry_name)

    def load_one_grid(self, hdf5_path: str, entry_name: str) -> Data:
        """Loads one grid.

        Args:
            hdf5_path (str): .HDF5 file name.
            entry_name (str): Name of the entry.

        Returns:
            :class:`torch_geometric.data.data.Data`: item with tensors x, y if present, entry_names.
        """
        with h5py.File(hdf5_path, "r") as hdf5_file:
            grp = hdf5_file[entry_name]

            mapped_features_group = grp[gridstorage.MAPPED_FEATURES]

            feature_data = [mapped_features_group[feature_name][:] for feature_name in self.features if feature_name[0] != "_"]
            x = torch.tensor(np.expand_dims(np.array(feature_data), axis=0), dtype=torch.float)

            # target
            if self.target is None:
                y = None
            elif targets.VALUES in grp and self.target in grp[targets.VALUES]:
                y = torch.tensor([grp[targets.VALUES][self.target][()]], dtype=torch.float)

                if self.task == targets.REGRESS and self.target_transform is True:
                    y = torch.sigmoid(torch.log(y))
                elif self.task is not targets.REGRESS and self.target_transform is True:
                    msg = f'Sigmoid transformation not possible for {self.task} tasks. Please change `task` to "regress" or set `target_transform` to `False`.'
                    raise ValueError(msg)
            else:
                y = None
                possible_targets = grp[targets.VALUES].keys()
                if self.train_source is None:
                    msg = (
                        f"Target {self.target} missing in entry {entry_name} in file {hdf5_path}, possible targets are {possible_targets}.\n\t"
                        "Use the query class to add more target values to input data."
                    )
                    raise ValueError(msg)

        # Wrap up the data in this object, for the collate_fn to handle it properly:
        data = Data(x=x, y=y)
        data.entry_names = entry_name

        return data


class GraphDataset(DeeprankDataset):
    """Class to load the .HDF5 files data into graphs.

    Args:
        hdf5_path (str | list): Path to .HDF5 file(s). For multiple .HDF5 files, insert the paths in a list. Defaults to None.
        subset (list[str] | None, optional): list of keys from .HDF5 file to include. Defaults to None (meaning include all).
        train_source (str | class:`GraphDataset` | None, optional): data to inherit information from the training dataset or the pre-trained model.
            If None, the current dataset is considered as the training set. Otherwise, `train_source` needs to be a dataset of the same class or
            the path of a DeepRank2 pre-trained model. If set, the parameters `node_features`, `edge_features`, `features_transform`,
            `target`, `target_transform`, `task`, and `classes` will be inherited from `train_source`. If standardization was performed in the
            training dataset/step, also the attributes `means` and `devs` will be inherited from `train_source`, and they will be used to scale
            the validation/testing set.
            Defaults to None.
        node_features (list[str] | str | Literal["all"] | None, optional): Consider all pre-computed node features ("all") or
            some defined node features (provide a list, example: ["res_type", "polarity", "bsa"]).
            The complete list can be found in `deeprank2.domain.nodestorage`.
            Value will be ignored and inherited from `train_source` if `train_source` is assigned.
            Defaults to "all".
        edge_features (list[str] | str | Literal["all"] | None, optional): Consider all pre-computed edge features ("all") or
            some defined edge features (provide a list, example: ["dist", "coulomb"]).
            The complete list can be found in `deeprank2.domain.edgestorage`.
            Value will be ignored and inherited from `train_source` if `train_source` is assigned.
            Defaults to "all".
        features_transform (dict | None, optional): Dictionary to indicate the transformations to apply to each feature in the dictionary, being the
            transformations lambda functions and/or standardization.
            Example: `features_transform = {'bsa': {'transform': lambda t:np.log(t+1),' standardize': True}}` for the feature `bsa`.
            An `all` key can be set in the dictionary for indicating to apply the same `standardize` and `transform` to all the features.
            Example: `features_transform = {'all': {'transform': lambda t:np.log(t+1), 'standardize': True}}`.
            If both `all` and feature name/s are present, the latter have the priority over what indicated in `all`.
            Value will be ignored and inherited from `train_source` if `train_source` is assigned.
            Defaults to None.
        clustering_method (str | None, optional): "mcl" for Markov cluster algorithm (see https://micans.org/mcl/),
            or "louvain" for Louvain method (see https://en.wikipedia.org/wiki/Louvain_method).
            In both options, for each graph, the chosen method first finds communities (clusters) of nodes and generates
            a torch tensor whose elements represent the cluster to which the node belongs to. Each tensor is then saved in
            the .HDF5 file as a :class:`Dataset` called "depth_0". Then, all cluster members beloging to the same community are
            pooled into a single node, and the resulting tensor is used to find communities among the pooled clusters.
            The latter tensor is saved into the .HDF5 file as a :class:`Dataset` called "depth_1". Both "depth_0" and "depth_1"
            :class:`Datasets` belong to the "cluster" Group. They are saved in the .HDF5 file to make them available to networks
            that make use of clustering methods. Defaults to None.
        target (str | None, optional): Default options are irmsd, lrmsd, fnat, binary, capri_class, and dockq.
            It can also be a custom-defined target given to the Query class as input (see: `deeprank2.query`);
            in this case, the task parameter needs to be explicitly specified as well.
            Only numerical target variables are supported, not categorical.
            If the latter is your case, please convert the categorical classes into
            numerical class indices before defining the :class:`GraphDataset` instance.
            Value will be ignored and inherited from `train_source` if `train_source` is assigned.
            Defaults to None.
        target_transform (bool, optional): Apply a log and then a sigmoid transformation to the target (for regression only).
            This puts the target value between 0 and 1, and can result in a more uniform target distribution and speed up the optimization.
            Value will be ignored and inherited from `train_source` if `train_source` is assigned.
            Defaults to False.
        target_filter (dict[str, str] | None, optional): Dictionary of type [target: cond] to filter the molecules.
            Note that the you can filter on a different target than the one selected as the dataset target.
            Defaults to None.
        task (Literal["regress", "classif"] | None, optional): 'regress' for regression or 'classif' for classification. Required if target not in
            ['irmsd', 'lrmsd', 'fnat', 'binary', 'capri_class', or 'dockq'], otherwise this setting is ignored.
            Automatically set to 'classif' if the target is 'binary' or 'capri_classes'.
            Automatically set to 'regress' if the target is 'irmsd', 'lrmsd', 'fnat', or 'dockq'.
            Value will be ignored and inherited from `train_source` if `train_source` is assigned.
            Defaults to None.
        classes (list[str] | list[int] | list[float] | None): Define the dataset target classes in classification mode.
            Value will be ignored and inherited from `train_source` if `train_source` is assigned.
            Defaults to None.
        use_tqdm (bool, optional): Show progress bar. Defaults to True.
        root (str, optional): Root directory where the dataset should be saved. Defaults to "./".
        check_integrity (bool, optional): Whether to check the integrity of the hdf5 files.
            Defaults to True.
    """

    def __init__(  # noqa: C901
        self,
        hdf5_path: str | list,
        subset: list[str] | None = None,
        train_source: str | GridDataset | None = None,
        node_features: list[str] | str | None = "all",
        edge_features: list[str] | str | None = "all",
        features_transform: dict | None = None,
        clustering_method: str | None = None,
        target: str | None = None,
        target_transform: bool = False,
        target_filter: dict[str, str] | None = None,
        task: Literal["regress", "classif"] | None = None,
        classes: list[str] | list[int] | list[float] | None = None,
        use_tqdm: bool = True,
        root: str = "./",
        check_integrity: bool = True,
    ):
        super().__init__(
            hdf5_path,
            subset,
            train_source,
            target,
            target_transform,
            target_filter,
            task,
            classes,
            use_tqdm,
            root,
            check_integrity,
        )

        self.default_vars = {k: v.default for k, v in inspect.signature(self.__init__).parameters.items() if v.default is not inspect.Parameter.empty}
        self.default_vars["classes_to_index"] = None
        self.node_features = node_features
        self.edge_features = edge_features
        self.clustering_method = clustering_method
        self.target_transform = target_transform
        self.features_transform = features_transform

        if train_source is not None:
            self.inherited_params = [
                "node_features",
                "edge_features",
                "features_transform",
                "target",
                "target_transform",
                "task",
                "classes",
                "classes_to_index",
            ]
            self._check_and_inherit_train(GraphDataset, self.inherited_params)
            self._check_features()

        else:
            self._check_features()
            self.inherited_params = None

            try:
                fname, mol = self.index_entries[0]
            except IndexError as e:
                msg = "No entries found in the dataset. Please check the dataset parameters."
                raise IndexError(msg) from e
            with h5py.File(fname, "r") as f5:
                grp = f5[mol]
                possible_targets = grp[targets.VALUES].keys()
                if self.target is None:
                    msg = f"Please set the target during training dataset definition; targets present in the file/s are {possible_targets}."
                    raise ValueError(msg)
                if self.target not in possible_targets:
                    msg = f"Target {self.target} not present in the file/s; targets present in the file/s are {possible_targets}."
                    raise ValueError(msg)

        self.features_dict = {}
        self.features_dict[Nfeat.NODE] = self.node_features
        self.features_dict[Efeat.EDGE] = self.edge_features
        if self.target is not None:
            if isinstance(self.target, str):
                self.features_dict[targets.VALUES] = [self.target]
            else:
                self.features_dict[targets.VALUES] = self.target

        standardize = False
        if self.features_transform:
            standardize = any(self.features_transform[key].get("standardize") for key, _ in self.features_transform.items())

        if standardize and (train_source is None):
            if self.means or self.devs is None:
                if self.df is None:
                    self.hdf5_to_pandas()
                self._compute_mean_std()
        elif standardize and (train_source is not None):
            self.means = self.train_means
            self.devs = self.train_devs

    def get(self, idx: int) -> Data:
        """Gets one graph item from its unique index.

        Args:
            idx (int): Index of the item, ranging from 0 to len(dataset).

        Returns:
            :class:`torch_geometric.data.data.Data`: item with tensors x, y if present, edge_index, edge_attr, pos, entry_names.
        """
        fname, mol = self.index_entries[idx]
        return self.load_one_graph(fname, mol)

    def load_one_graph(self, fname: str, entry_name: str) -> Data:  # noqa: PLR0915, C901
        """Loads one graph.

        Args:
            fname (str): .HDF5 file name.
            entry_name (str): Name of the entry.

        Returns:
            :class:`torch_geometric.data.data.Data`: item with tensors x, y if present, edge_index, edge_attr, pos, entry_names.
        """
        with h5py.File(fname, "r") as f5:
            grp = f5[entry_name]

            # node features
            if len(self.node_features) > 0:
                node_data = ()
                for feat in self.node_features:
                    # resetting transformation and standardization for each feature
                    transform = None
                    standard = None

                    if feat[0] != "_":  # ignore metafeatures
                        vals = grp[f"{Nfeat.NODE}/{feat}"][()]
                        # get feat transformation and standardization
                        if self.features_transform is not None:
                            transform = self.features_transform.get("all", {}).get("transform")
                            standard = self.features_transform.get("all", {}).get("standardize")
                            # if no transformation is set for all features, check if one is set for the current feature
                            if (transform is None) and (feat in self.features_transform):
                                transform = self.features_transform.get(feat, {}).get("transform")
                            # if no standardization is set for all features, check if one is set for the current feature
                            if (standard is None) and (feat in self.features_transform):
                                standard = self.features_transform.get(feat, {}).get("standardize")

                        # apply transformation
                        if transform:
                            with warnings.catch_warnings(record=True) as w:
                                vals = transform(vals)
                                if len(w) > 0:
                                    msg = (
                                        f"Invalid value occurs in {entry_name}, file {fname},when applying {transform} for feature {feat}.\n\t"
                                        f"Please change the transformation function for {feat}."
                                    )
                                    raise ValueError(msg)

                        if vals.ndim == 1:  # features with only one channel
                            vals = vals.reshape(-1, 1)
                            if standard:
                                vals = (vals - self.means[feat]) / self.devs[feat]
                        elif standard:
                            reshaped_mean = [mean_value for mean_key, mean_value in self.means.items() if feat in mean_key]
                            reshaped_dev = [dev_value for dev_key, dev_value in self.devs.items() if feat in dev_key]
                            vals = (vals - reshaped_mean) / reshaped_dev
                        node_data += (vals,)
                x = torch.tensor(np.hstack(node_data), dtype=torch.float)
            else:
                x = None
                _log.warning("No node features set.")

            # edge index,
            # we have to have all the edges i.e : (i,j) and (j,i)
            if Efeat.INDEX in grp[Efeat.EDGE]:
                ind = grp[f"{Efeat.EDGE}/{Efeat.INDEX}"][()]
                if ind.ndim == 2:  # noqa: PLR2004
                    ind = np.vstack((ind, np.flip(ind, 1))).T
                edge_index = torch.tensor(ind, dtype=torch.long).contiguous()
            else:
                edge_index = torch.empty((2, 0), dtype=torch.long)

            # edge feature
            # we have to have all the edges i.e : (i,j) and (j,i)
            if len(self.edge_features) > 0:
                edge_data = ()
                for feat in self.edge_features:
                    # resetting transformation and standardization for each feature
                    transform = None
                    standard = None

                    if feat[0] != "_":  # ignore metafeatures
                        vals = grp[f"{Efeat.EDGE}/{feat}"][()]
                        # get feat transformation and standardization
                        if self.features_transform is not None:
                            transform = self.features_transform.get("all", {}).get("transform")
                            standard = self.features_transform.get("all", {}).get("standardize")
                            # if no transformation is set for all features, check if one is set for the current feature
                            if (transform is None) and (feat in self.features_transform):
                                transform = self.features_transform.get(feat, {}).get("transform")
                            # if no standardization is set for all features, check if one is set for the current feature
                            if (standard is None) and (feat in self.features_transform):
                                standard = self.features_transform.get(feat, {}).get("standardize")

                        # apply transformation
                        if transform:
                            with warnings.catch_warnings(record=True) as w:
                                vals = transform(vals)
                                if len(w) > 0:
                                    msg = (
                                        f"Invalid value occurs in {entry_name}, file {fname}, when applying {transform} for feature {feat}.\n\t"
                                        f"Please change the transformation function for {feat}."
                                    )
                                    raise ValueError(msg)

                        if vals.ndim == 1:
                            vals = vals.reshape(-1, 1)
                            if standard:
                                vals = (vals - self.means[feat]) / self.devs[feat]
                        elif standard:
                            reshaped_mean = [mean_value for mean_key, mean_value in self.means.items() if feat in mean_key]
                            reshaped_dev = [dev_value for dev_key, dev_value in self.devs.items() if feat in dev_key]
                            vals = (vals - reshaped_mean) / reshaped_dev
                        edge_data += (vals,)
                edge_data = np.hstack(edge_data)
                edge_data = np.vstack((edge_data, edge_data))
                edge_attr = torch.tensor(edge_data, dtype=torch.float).contiguous()
            else:
                edge_attr = torch.empty((edge_index.shape[1], 0), dtype=torch.float).contiguous()

            # target
            if self.target is None:
                y = None
            elif targets.VALUES in grp and self.target in grp[targets.VALUES]:
                y = torch.tensor([grp[f"{targets.VALUES}/{self.target}"][()]], dtype=torch.float).contiguous()

                if self.task == targets.REGRESS and self.target_transform is True:
                    y = torch.sigmoid(torch.log(y))
                elif self.task is not targets.REGRESS and self.target_transform is True:
                    msg = f'Sigmoid transformation not possible for {self.task} tasks. Please change `task` to "regress" or set `target_transform` to `False`.'
                    raise ValueError(msg)

            else:
                y = None
                possible_targets = grp[targets.VALUES].keys()
                if self.train_source is None:
                    msg = (
                        f"Target {self.target} missing in entry {entry_name} in file {fname}, possible targets are {possible_targets}.\n\t"
                        "Use the query class to add more target values to input data."
                    )
                    raise ValueError(msg)

            # positions
            pos = torch.tensor(grp[f"{Nfeat.NODE}/{Nfeat.POSITION}/"][()], dtype=torch.float).contiguous()

            # cluster
            cluster0 = None
            cluster1 = None
            if self.clustering_method is not None and "clustering" in grp:
                if self.clustering_method in grp["clustering"]:
                    if "depth_0" in grp[f"clustering/{self.clustering_method}"] and "depth_1" in grp[f"clustering/{self.clustering_method}"]:
                        cluster0 = torch.tensor(
                            grp["clustering/" + self.clustering_method + "/depth_0"][()],
                            dtype=torch.long,
                        )
                        cluster1 = torch.tensor(
                            grp["clustering/" + self.clustering_method + "/depth_1"][()],
                            dtype=torch.long,
                        )
                    else:
                        _log.warning("no clusters detected")
                else:
                    _log.warning(f"no clustering/{self.clustering_method} detected")

        # load
        data = Data(x=x, edge_index=edge_index, edge_attr=edge_attr, y=y, pos=pos)

        data.cluster0 = cluster0
        data.cluster1 = cluster1

        data.entry_names = entry_name

        return data

    def _check_features(self) -> None:  # noqa: C901
        """Checks if the required features exist."""
        f = h5py.File(self.hdf5_paths[0], "r")
        mol_key = next(iter(f.keys()))

        # read available node features
        self.available_node_features = list(f[f"{mol_key}/{Nfeat.NODE}/"].keys())
        self.available_node_features = [key for key in self.available_node_features if key[0] != "_"]  # ignore metafeatures

        # read available edge features
        self.available_edge_features = list(f[f"{mol_key}/{Efeat.EDGE}/"].keys())
        self.available_edge_features = [key for key in self.available_edge_features if key[0] != "_"]  # ignore metafeatures

        f.close()

        # check node features
        missing_node_features = []
        if self.node_features == "all":
            self.node_features = self.available_node_features
            self.default_vars["node_features"] = self.node_features
        else:
            if not isinstance(self.node_features, list):
                if self.node_features is None:
                    self.node_features = []
                else:
                    self.node_features = [self.node_features]
            for feat in self.node_features:
                if feat not in self.available_node_features:
                    _log.info(f"The node feature _{feat}_ was not found in the file {self.hdf5_paths[0]}.")
                    missing_node_features.append(feat)

        # check edge features
        missing_edge_features = []
        if self.edge_features == "all":
            self.edge_features = self.available_edge_features
            self.default_vars["edge_features"] = self.edge_features
        else:
            if not isinstance(self.edge_features, list):
                if self.edge_features is None:
                    self.edge_features = []
                else:
                    self.edge_features = [self.edge_features]
            for feat in self.edge_features:
                if feat not in self.available_edge_features:
                    _log.info(f"The edge feature _{feat}_ was not found in the file {self.hdf5_paths[0]}.")
                    missing_edge_features.append(feat)

        # raise error if any features are missing
        if missing_node_features + missing_edge_features:
            miss_node_error, miss_edge_error = "", ""
            _log.info(
                "\nCheck feature_modules passed to the preprocess function.\
                Probably, the feature wasn't generated during the preprocessing step.",
            )
            if missing_node_features:
                _log.info(f"\nAvailable node features: {self.available_node_features}\n")
                miss_node_error = f"\nMissing node features: {missing_node_features} \
                                    \nAvailable node features: {self.available_node_features}"
            if missing_edge_features:
                _log.info(f"\nAvailable edge features: {self.available_edge_features}\n")
                miss_edge_error = f"\nMissing edge features: {missing_edge_features} \
                                    \nAvailable edge features: {self.available_edge_features}"
            msg = (
                f"Not all features could be found in the file {self.hdf5_paths[0]}.\n\t"
                "Check feature_modules passed to the preprocess function.\n\t"
                "Probably, the feature wasn't generated during the preprocessing step.\n\t"
                f"{miss_node_error}{miss_edge_error}"
            )
            raise ValueError(msg)


def save_hdf5_keys(
    f_src_path: str,
    src_ids: list[str],
    f_dest_path: str,
    hardcopy: bool = False,
) -> None:
    """Save references to keys in src_ids in a new .HDF5 file.

    Args:
        f_src_path (str): The path to the .HDF5 file containing the keys.
        src_ids (list[str]): Keys to be saved in the new .HDF5 file. It should be a list containing at least one key.
        f_dest_path (str): The path to the new .HDF5 file.
        hardcopy (bool, optional): If False, the new file contains only references (external links, see :class:`ExternalLink` class from `h5py`)
            to the original .HDF5 file.
            If True, the new file contains a copy of the objects specified in src_ids (see h5py :class:`HardLink` from `h5py`).
            Defaults to False.
    """
    if not all(isinstance(d, str) for d in src_ids):
        msg = "data_ids should be a list containing strings."
        raise TypeError(msg)

    with h5py.File(f_dest_path, "w") as f_dest, h5py.File(f_src_path, "r") as f_src:
        for key in src_ids:
            if hardcopy:
                f_src.copy(f_src[key], f_dest)
            else:
                f_dest[key] = h5py.ExternalLink(f_src_path, "/" + key)



## main group
EDGE = "edge_features"

## metafeatures
NAME = "_name"
INDEX = "_index"

## generic features
DISTANCE = "distance"  # float; former FEATURENAME_EDGEDISTANCE
SAMECHAIN = "same_chain"  # bool; former FEATURE_EDGE_INTERACTIONTYPE & FEATURENAME_EDGESAMECHAIN
SAMERES = "same_res"  # bool

## interactions
COVALENT = "covalent"  # bool; former FEATURENAME_COVALENT
ELEC = "electrostatic"  # float; former FEATURENAME_EDGECOULOMB
VDW = "vanderwaals"  # float; former FEATURENAME_EDGEVANDERWAALS



import numpy as np

from deeprank2.domain import nodestorage as Nfeat
from deeprank2.domain.aminoacidlist import amino_acids
from deeprank2.molstruct.atom import Atom
from deeprank2.molstruct.residue import Residue, SingleResidueVariant
from deeprank2.utils.graph import Graph


def add_features(  # noqa:D103
    pdb_path: str,  # noqa: ARG001
    graph: Graph,
    single_amino_acid_variant: SingleResidueVariant | None = None,
) -> None:
    profile_amino_acid_order = sorted(amino_acids, key=lambda aa: aa.three_letter_code)

    for node in graph.nodes:
        if isinstance(node.id, Residue):
            residue = node.id
        elif isinstance(node.id, Atom):
            atom = node.id
            residue = atom.residue
        else:
            msg = f"Unexpected node type: {type(node.id)}"
            raise TypeError(msg)

        pssm_row = residue.get_pssm()
        profile = np.array([pssm_row.get_conservation(amino_acid) for amino_acid in profile_amino_acid_order])
        node.features[Nfeat.PSSM] = profile
        node.features[Nfeat.INFOCONTENT] = pssm_row.information_content

        if single_amino_acid_variant is not None:
            if residue == single_amino_acid_variant.residue:
                # only the variant residue can have a variant and wildtype amino acid
                conservation_wildtype = pssm_row.get_conservation(single_amino_acid_variant.wildtype_amino_acid)
                conservation_variant = pssm_row.get_conservation(single_amino_acid_variant.variant_amino_acid)
                node.features[Nfeat.CONSERVATION] = conservation_wildtype
                node.features[Nfeat.DIFFCONSERVATION] = conservation_variant - conservation_wildtype
            else:
                # all nodes must have the same features, so set them to zero here
                node.features[Nfeat.CONSERVATION] = 0.0
                node.features[Nfeat.DIFFCONSERVATION] = 0.0



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

import h5py
import numpy as np
import pdb2sql.transform
from scipy.spatial import distance_matrix

from deeprank2.domain import edgestorage as Efeat
from deeprank2.domain import nodestorage as Nfeat
from deeprank2.domain import targetstorage as targets
from deeprank2.molstruct.atom import Atom
from deeprank2.molstruct.pair import AtomicContact, Contact, ResidueContact
from deeprank2.molstruct.residue import Residue
from deeprank2.utils.grid import Augmentation, Grid, GridSettings, MapMethod

if TYPE_CHECKING:
    from collections.abc import Callable

    from numpy.typing import NDArray

_log = logging.getLogger(__name__)


class Edge:
    """Graph edge."""

    def __init__(self, id_: Contact):
        self.id = id_
        self.features = {}

    def add_feature(self, feature_name: str, feature_function: Callable[[Contact], float]) -> None:
        feature_value = feature_function(self.id)

        self.features[feature_name] = feature_value

    @property
    def position1(self) -> np.array:
        return self.id.item1.position

    @property
    def position2(self) -> np.array:
        return self.id.item2.position

    def has_nan(self) -> bool:
        """Whether there are any NaN values in the edge's features."""
        return any(np.any(np.isnan(feature_data)) for feature_data in self.features.values())


class Node:
    """Graph node."""

    def __init__(self, id_: Atom | Residue):
        if isinstance(id_, Atom):
            self._type = "atom"
        elif isinstance(id_, Residue):
            self._type = "residue"
        else:
            raise TypeError(type(id_))

        self.id = id_
        self.features = {}

    @property
    def type(self) -> str:
        return self._type

    def has_nan(self) -> bool:
        """Whether there are any NaN values in the node's features."""
        return any(np.any(np.isnan(feature_data)) for feature_data in self.features.values())

    def add_feature(
        self,
        feature_name: str,
        feature_function: Callable[[Atom | Residue], NDArray],
    ) -> None:
        feature_value = feature_function(self.id)

        if len(feature_value.shape) != 1:
            shape_s = "x".join(feature_value.shape)
            msg = f"Expected a 1-dimensional array for feature {feature_name}, but got {shape_s}"
            raise ValueError(msg)

        self.features[feature_name] = feature_value

    @property
    def position(self) -> np.array:
        return self.id.position


class Graph:
    """Graph."""

    def __init__(self, id_: str):
        self.id = id_

        self._nodes = {}
        self._edges = {}

        # targets are optional and may be set later
        self.targets = {}

        # the center only needs to be set when this graph should be mapped to a grid.
        self.center = np.array((0.0, 0.0, 0.0))

    def add_node(self, node: Node) -> None:
        self._nodes[node.id] = node

    def get_node(self, id_: Atom | Residue) -> Node:
        return self._nodes[id_]

    def add_edge(self, edge: Edge) -> None:
        self._edges[edge.id] = edge

    def get_edge(self, id_: Contact) -> Edge:
        return self._edges[id_]

    @property
    def nodes(self) -> list[Node]:
        return list(self._nodes.values())

    @property
    def edges(self) -> list[Node]:
        return list(self._edges.values())

    def has_nan(self) -> bool:
        """Whether there are any NaN values in the graph's features."""
        for node in self._nodes.values():
            if node.has_nan():
                return True
        return any(edge.has_nan() for edge in self._edges.values())

    def _map_point_features(
        self,
        grid: Grid,
        method: MapMethod,
        feature_name: str,
        points: list[NDArray],
        values: list[float | NDArray],
        augmentation: Augmentation | None = None,
    ) -> None:
        points = np.stack(points, axis=0)

        if augmentation is not None:
            points = pdb2sql.transform.rot_xyz_around_axis(
                points,
                augmentation.axis,
                augmentation.angle,
                self.center,
            )

        for point_index in range(points.shape[0]):
            position = points[point_index]
            value = values[point_index]

            grid.map_feature(position, feature_name, value, method)

    def map_to_grid(
        self,
        grid: Grid,
        method: MapMethod,
        augmentation: Augmentation | None = None,
    ) -> None:
        # order edge features by xyz point
        points = []
        feature_values = {}
        for edge in self._edges.values():
            points += [edge.position1, edge.position2]

            for feature_name, feature_value in edge.features.items():
                feature_values[feature_name] = feature_values.get(feature_name, []) + [  # noqa: RUF005
                    feature_value,
                    feature_value,
                ]

        # map edge features to grid
        for feature_name, values in feature_values.items():
            self._map_point_features(
                grid,
                method,
                feature_name,
                points,
                values,
                augmentation,
            )

        # order node features by xyz point
        points = []
        feature_values = {}
        for node in self._nodes.values():
            points.append(node.position)

            for feature_name, feature_value in node.features.items():
                feature_values[feature_name] = feature_values.get(feature_name, []) + [feature_value]  # noqa: RUF005

        # map node features to grid
        for feature_name, values in feature_values.items():
            self._map_point_features(
                grid,
                method,
                feature_name,
                points,
                values,
                augmentation,
            )

    def write_to_hdf5(self, hdf5_path: str) -> None:
        """Write a featured graph to an hdf5 file, according to deeprank standards."""
        with h5py.File(hdf5_path, "a") as hdf5_file:
            # create groups to hold data
            graph_group = hdf5_file.require_group(self.id)
            node_features_group = graph_group.create_group(Nfeat.NODE)
            edge_feature_group = graph_group.create_group(Efeat.EDGE)

            # store node names and chain_ids
            node_names = np.array([str(key) for key in self._nodes]).astype("S")
            node_features_group.create_dataset(Nfeat.NAME, data=node_names)
            chain_ids = np.array([str(key).split()[1] for key in self._nodes]).astype("S")
            node_features_group.create_dataset(Nfeat.CHAINID, data=chain_ids)

            # store node features
            node_key_list = list(self._nodes.keys())
            first_node_data = next(iter(self._nodes.values())).features
            node_feature_names = list(first_node_data.keys())
            for node_feature_name in node_feature_names:
                node_feature_data = [node.features[node_feature_name] for node in self._nodes.values()]

                node_features_group.create_dataset(node_feature_name, data=node_feature_data)

            # identify edges
            edge_indices = []
            edge_names = []

            first_edge_data = next(iter(self._edges.values())).features
            edge_feature_names = list(first_edge_data.keys())

            edge_feature_data = {name: [] for name in edge_feature_names}

            for edge_id, edge in self._edges.items():
                id1, id2 = edge_id
                node_index1 = node_key_list.index(id1)
                node_index2 = node_key_list.index(id2)

                edge_indices.append((node_index1, node_index2))
                edge_names.append(f"{id1}-{id2}")

                for edge_feature_name in edge_feature_names:
                    edge_feature_data[edge_feature_name].append(edge.features[edge_feature_name])

            # store edge names and indices
            edge_feature_group.create_dataset(Efeat.NAME, data=np.array(edge_names).astype("S"))
            edge_feature_group.create_dataset(Efeat.INDEX, data=edge_indices)

            # store edge features
            for edge_feature_name in edge_feature_names:
                edge_feature_group.create_dataset(edge_feature_name, data=edge_feature_data[edge_feature_name])

            # store target values
            score_group = graph_group.create_group(targets.VALUES)
            for target_name, target_data in self.targets.items():
                score_group.create_dataset(target_name, data=target_data)

    @staticmethod
    def _find_unused_augmentation_name(unaugmented_id: str, hdf5_path: str) -> str:
        prefix = f"{unaugmented_id}_"

        if os.path.isfile(hdf5_path):
            with h5py.File(hdf5_path, "r") as hdf5_file:
                entry_names_taken = [entry_name for entry_name in hdf5_file if entry_name.startswith(prefix)]

        augmentation_count = 0
        chosen_name = f"{prefix}{augmentation_count:03}"
        while chosen_name in entry_names_taken:
            augmentation_count += 1
            chosen_name = f"{prefix}{augmentation_count:03}"

        return chosen_name

    def write_as_grid_to_hdf5(
        self,
        hdf5_path: str,
        settings: GridSettings,
        method: MapMethod,
        augmentation: Augmentation | None = None,
    ) -> str:
        id_ = self.id
        if augmentation is not None:
            id_ = self._find_unused_augmentation_name(id_, hdf5_path)

        grid = Grid(id_, self.center.tolist(), settings)

        self.map_to_grid(grid, method, augmentation)
        grid.to_hdf5(hdf5_path)

        # store target values
        with h5py.File(hdf5_path, "a") as hdf5_file:
            grp = hdf5_file[id_]

            targets_group = grp.require_group(targets.VALUES)
            for target_name, target_data in self.targets.items():
                if target_name not in targets_group:
                    targets_group.create_dataset(target_name, data=target_data)
                else:
                    targets_group[target_name][()] = target_data

        return hdf5_path

    def get_all_chains(self) -> list[str]:
        if isinstance(self.nodes[0].id, Residue):
            chains = {str(res.chain).split()[1] for res in [node.id for node in self.nodes]}
        elif isinstance(self.nodes[0].id, Atom):
            chains = {str(res.chain).split()[1] for res in [node.id.residue for node in self.nodes]}
        else:
            return None
        return list(chains)

    @staticmethod
    def build_graph(
        nodes: list[Atom] | list[Residue],
        graph_id: str,
        max_edge_length: float,
    ) -> Graph:
        """Builds a graph.

        Args:
            nodes (list[Atom] | list[Residue]): List of `Atom`s or `Residue`s to include in graph.
                All nodes must be of same type.
            graph_id (str): Human readable identifier for graph.
            max_edge_length (float): Maximum distance between two nodes to connect them with an edge.

        Returns:
            Graph: Containing nodes (with positions) and edges.

        Raises:
            TypeError: if `nodes` argument contains a mix of different types.
        """
        if all(isinstance(node, Atom) for node in nodes):
            atoms = nodes
            NodeContact = AtomicContact
        elif all(isinstance(node, Residue) for node in nodes):
            # collect the set of atoms and remember which are on the same residue (by index)
            atoms = []
            atoms_residues = []
            for residue_index, residue in enumerate(nodes):
                for atom in residue.atoms:
                    atoms.append(atom)
                    atoms_residues.append(residue_index)
            atoms_residues = np.array(atoms_residues)
            NodeContact = ResidueContact
        else:
            msg = "All nodes in the graph must be of the same type."
            raise TypeError(msg)

        positions = np.empty((len(atoms), 3))
        for atom_index, atom in enumerate(atoms):
            positions[atom_index] = atom.position
        neighbours = max_edge_length > distance_matrix(positions, positions, p=2)

        index_pairs = np.transpose(np.nonzero(neighbours))  # atom pairs
        if NodeContact == ResidueContact:
            index_pairs = np.unique(atoms_residues[index_pairs], axis=0)  # residue pairs

        graph = Graph(graph_id)

        for index1, index2 in index_pairs:
            if index1 != index2:
                node1 = Node(nodes[index1])
                node2 = Node(nodes[index2])
                contact = NodeContact(node1.id, node2.id)

                node1.features[Nfeat.POSITION] = node1.id.position
                node2.features[Nfeat.POSITION] = node2.id.position

                # The same node will be added multiple times, but the Graph class fixes this.
                graph.add_node(node1)
                graph.add_node(node2)
                graph.add_edge(Edge(contact))

        return graph



import logging
import os

import numpy as np
from pdb2sql import interface as pdb2sql_interface
from pdb2sql import pdb2sql as pdb2sql_object
from scipy.spatial import distance_matrix

from deeprank2.domain.aminoacidlist import amino_acids_by_code
from deeprank2.molstruct.atom import Atom, AtomicElement
from deeprank2.molstruct.pair import Pair
from deeprank2.molstruct.residue import Residue
from deeprank2.molstruct.structure import Chain, PDBStructure

_log = logging.getLogger(__name__)


def _add_atom_to_residue(atom: Atom, residue: Residue) -> None:
    """Adds an `Atom` to a `Residue` if not already there.

    If no matching atom is found, add the current atom to the residue.
    If there's another atom with the same name, choose the one with the highest occupancy.
    """
    for other_atom in residue.atoms:
        if other_atom.name == atom.name and other_atom.occupancy < atom.occupancy:
            other_atom.change_altloc(atom)
            return
    residue.add_atom(atom)


def _add_atom_data_to_structure(
    structure: PDBStructure,
    pdb_obj: pdb2sql_object,
    **kwargs,  # noqa: ANN003
) -> None:
    """This subroutine retrieves pdb2sql atomic data for `PDBStructure` objects as defined in DeepRank2.

    This function should be called for one atom at a time.

    Args:
        structure (:class:`PDBStructure`): The structure to which this atom should be added to.
        pdb_obj (pdb2sql_object): The `pdb2sql` object to retrieve the data from.
        kwargs: as required by the get function for the `pdb2sql` object.
    """
    pdb2sql_columns = "x,y,z,name,altLoc,occ,element,chainID,resSeq,resName,iCode"
    data_keys = pdb2sql_columns.split(sep=",")
    for data_values in pdb_obj.get(pdb2sql_columns, **kwargs):
        atom_data = dict(zip(data_keys, data_values, strict=True))

        # exit function if this atom is already part of the structure
        if atom_data["altLoc"] not in (None, "", "A"):
            return

        atom_data["iCode"] = None if atom_data["iCode"] == "" else atom_data["iCode"]

        try:
            atom_data["aa"] = amino_acids_by_code[atom_data["resName"]]
        except KeyError:
            atom_data["aa"] = None
        atom_data["coordinates"] = np.array(data_values[:3])

        if not structure.has_chain(atom_data["chainID"]):
            structure.add_chain(Chain(structure, atom_data["chainID"]))
        chain = structure.get_chain(atom_data["chainID"])

        if not chain.has_residue(atom_data["resSeq"], atom_data["iCode"]):
            chain.add_residue(Residue(chain, atom_data["resSeq"], atom_data["aa"], atom_data["iCode"]))
        residue = chain.get_residue(atom_data["resSeq"], atom_data["iCode"])

        atom = Atom(
            residue,
            atom_data["name"],
            AtomicElement[atom_data["element"]],
            atom_data["coordinates"],
            atom_data["occ"],
        )
        _add_atom_to_residue(atom, residue)


def get_structure(pdb_obj: pdb2sql_object, id_: str) -> PDBStructure:
    """Builds a structure from rows in a pdb file.

    Args:
        pdb_obj (pdb2sql object): The pdb structure that we're investigating.
        id_ (str): Unique id for the pdb structure.

    Returns:
        PDBStructure: The structure object, giving access to chains, residues, atoms.
    """
    structure = PDBStructure(id_)
    _add_atom_data_to_structure(structure, pdb_obj, model=0)
    return structure


def get_contact_atoms(
    pdb_path: str,
    chain_ids: list[str],
    influence_radius: float,
) -> list[Atom]:
    """Gets the contact atoms from pdb2sql and wraps them in python objects."""
    interface = pdb2sql_interface(pdb_path)
    pdb_name = os.path.splitext(os.path.basename(pdb_path))[0]
    structure = PDBStructure(f"contact_atoms_{pdb_name}")

    try:
        atom_indexes = interface.get_contact_atoms(
            cutoff=influence_radius,
            chain1=chain_ids[0],
            chain2=chain_ids[1],
        )
        pdb_rowID = atom_indexes[chain_ids[0]] + atom_indexes[chain_ids[1]]
        _add_atom_data_to_structure(structure, interface, rowID=pdb_rowID)
    finally:
        interface._close()  # noqa: SLF001

    return structure.get_atoms()


def get_residue_contact_pairs(
    pdb_path: str,
    structure: PDBStructure,
    chain_id1: str,
    chain_id2: str,
    influence_radius: float,
) -> list[Pair]:
    """Find all residue pairs that may influence each other.

    Args:
        pdb_path (str): The path of the pdb file, that the structure was built from.
        structure (:class:`PDBStructure`): From which to take the residues.
        chain_id1 (str): First protein chain identifier.
        chain_id2 (str): Second protein chain identifier.
        influence_radius (float): Maximum distance between residues to consider them as interacting.

    Returns:
        list[Pair]: The pairs of contacting residues.
    """
    # Find out which residues are pairs
    interface = pdb2sql_interface(pdb_path)
    try:
        contact_residues = interface.get_contact_residues(
            cutoff=influence_radius,
            chain1=chain_id1,
            chain2=chain_id2,
            return_contact_pairs=True,
        )
    finally:
        interface._close()  # noqa: SLF001

    # Map to residue objects
    residue_pairs = set()
    for residue_key1, residue_contacts in contact_residues.items():
        residue1 = _get_residue_from_key(structure, residue_key1)
        for residue_key2 in residue_contacts:
            residue2 = _get_residue_from_key(structure, residue_key2)
            residue_pairs.add(Pair(residue1, residue2))

    return residue_pairs


def _get_residue_from_key(
    structure: PDBStructure,
    residue_key: tuple[str, int, str],
) -> Residue:
    """Returns a residue object given a pdb2sql-formatted residue key."""
    residue_chain_id, residue_number, residue_name = residue_key
    chain = structure.get_chain(residue_chain_id)

    for residue in chain.residues:
        if residue.number == residue_number and residue.amino_acid is not None and residue.amino_acid.three_letter_code == residue_name:
            return residue
    msg = f"Residue ({residue_key}) not found in {structure.id}."
    raise ValueError(msg)


def get_surrounding_residues(
    structure: Chain | PDBStructure,
    residue: Residue,
    radius: float,
) -> list[Residue]:
    """Get the residues that lie within a radius around a residue.

    Args:
        structure (:class:`Chain` | :class:`PDBStructure`): The structure to take residues from.
        residue (:class:`Residue`): The residue in the structure.
        radius (float): Max distance in Ångström between atoms of the residue and the other residues.

    Returns:
        list[:class:`Residue`]: The surrounding residues.
    """
    structure_atoms = structure.get_atoms()
    structure_atom_positions = [atom.position for atom in structure_atoms]
    residue_atom_positions = [atom.position for atom in residue.atoms]
    pairwise_distances = distance_matrix(
        structure_atom_positions,
        residue_atom_positions,
        p=2,
    )

    surrounding_residues = set()
    for structure_atom_index, structure_atom in enumerate(structure_atoms):
        shortest_distance = np.min(pairwise_distances[structure_atom_index, :])
        if shortest_distance < radius:
            surrounding_residues.add(structure_atom.residue)

    return list(surrounding_residues)



import copy
import inspect
import logging
import re
import warnings
from time import time
from typing import Any

import dill
import h5py
import numpy as np
import torch
import torch.nn.functional as F
from torch import nn
from torch_geometric.loader import DataLoader
from tqdm import tqdm

from deeprank2.dataset import GraphDataset, GridDataset
from deeprank2.domain import losstypes as losses
from deeprank2.domain import targetstorage as targets
from deeprank2.utils.community_pooling import community_detection, community_pooling
from deeprank2.utils.earlystopping import EarlyStopping
from deeprank2.utils.exporters import HDF5OutputExporter, OutputExporter, OutputExporterCollection

# ruff: noqa: PYI041 (usage depends on type in this module)
_log = logging.getLogger(__name__)


class Trainer:
    """Class from which the network is trained, evaluated and tested.

    Args:
        neuralnet (child class of :class:`torch.nn.Module`, optional): Neural network class (ex. :class:`GINet`, :class:`Foutnet` etc.).
            It should subclass :class:`torch.nn.Module`, and it shouldn't be specific to regression or classification
            in terms of output shape (:class:`Trainer` class takes care of formatting the output shape according to the task).
            More specifically, in classification task cases, softmax shouldn't be used as the last activation function.
            Defaults to None.
        dataset_train (:class:`GraphDataset` | :class:`GridDataset` | None, optional): Training set used during training.
            Can't be None if pretrained_model is also None. Defaults to None.
        dataset_val (:class:`GraphDataset` | :class:`GridDataset` | None, optional): Evaluation set used during training.
            If None, training set will be split randomly into training set and validation set during training, using val_size parameter.
            Defaults to None.
        dataset_test (:class:`GraphDataset` | :class:`GridDataset` | None, optional): Independent evaluation set. Defaults to None.
        val_size (float | int | None, optional): Fraction of dataset (if float) or number of datapoints (if int) to use for validation.
            Only used if dataset_val is not specified. Can be set to 0 if no validation set is needed. Defaults to None (in _divide_dataset function).
        test_size (float | int | None, optional): Fraction of dataset (if float) or number of datapoints (if int) to use for test dataset.
            Only used if dataset_test is not specified. Can be set to 0 if no test set is needed. Defaults to None.
        class_weights (bool, optional): Assign class weights based on the dataset content. Defaults to False.
        pretrained_model (str | None, optional): Path to pre-trained model. Defaults to None.
        cuda (bool, optional): Whether to use CUDA. Defaults to False.
        ngpu (int, optional): Number of GPU to be used. Defaults to 0.
        output_exporters (list[OutputExporter] | None, optional): The output exporters to use for saving/exploring/plotting predictions/targets/losses
            over the epochs. If None, defaults to :class:`HDF5OutputExporter`, which saves all the results in an .HDF5 file stored in ./output directory.
            Defaults to None.
    """

    def __init__(  # noqa: PLR0915, C901
        self,
        neuralnet: nn.Module = None,
        dataset_train: GraphDataset | GridDataset | None = None,
        dataset_val: GraphDataset | GridDataset | None = None,
        dataset_test: GraphDataset | GridDataset | None = None,
        val_size: float | int | None = None,
        test_size: float | int | None = None,
        class_weights: bool = False,
        pretrained_model: str | None = None,
        cuda: bool = False,
        ngpu: int = 0,
        output_exporters: list[OutputExporter] | None = None,
    ):
        self.neuralnet = neuralnet
        self.pretrained_model = pretrained_model

        self._init_datasets(dataset_train, dataset_val, dataset_test, val_size, test_size)

        self.cuda = cuda
        self.ngpu = ngpu

        if self.cuda and torch.cuda.is_available():
            self.device = torch.device("cuda")
            if self.ngpu == 0:
                self.ngpu = 1
                _log.info("CUDA detected. Setting number of GPUs to 1.")
        elif self.cuda and not torch.cuda.is_available():
            _log.error(
                """
                --> CUDA not detected: Make sure that CUDA is installed
                    and that you are running on GPUs.\n
                --> To turn CUDA off set cuda=False in Trainer.\n
                --> Aborting the experiment \n\n'
                """,
            )
            msg = """
                --> CUDA not detected: Make sure that CUDA is installed
                    and that you are running on GPUs.\n
                --> To turn CUDA off set cuda=False in Trainer.\n
                --> Aborting the experiment \n\n'
                """
            raise ValueError(msg)
        else:
            self.device = torch.device("cpu")
            if self.ngpu > 0:
                _log.error(
                    """
                    --> CUDA not detected.
                        Set cuda=True in Trainer to turn CUDA on.\n
                    --> Aborting the experiment \n\n
                    """,
                )
                msg = """
                    --> CUDA not detected.
                        Set cuda=True in Trainer to turn CUDA on.\n
                    --> Aborting the experiment \n\n
                    """
                raise ValueError(msg)

        _log.info(f"Device set to {self.device}.")
        if self.device.type == "cuda":
            _log.info(f"CUDA device name is {torch.cuda.get_device_name(0)}.")
            _log.info(f"Number of GPUs set to {self.ngpu}.")

        self._init_output_exporters(output_exporters)

        # other attributes not set in init
        self.data_type = None
        self.batch_size_train = None
        self.batch_size_test = None
        self.shuffle = None
        self.model_load_state_dict = None

        if self.pretrained_model is None:
            if self.dataset_train is None:
                msg = "No training data specified. Training data is required if there is no pretrained model."
                raise ValueError(msg)
            if self.neuralnet is None:
                msg = "No neural network specified. Specifying a model framework is required if there is no pretrained model."
                raise ValueError(msg)

            self._init_from_dataset(self.dataset_train)
            self.optimizer = None
            self.class_weights = class_weights
            self.subset = self.dataset_train.subset
            self.epoch_saved_model = None

            if self.target is None:
                msg = "No target set. You need to choose a target (set in the dataset) for training."
                raise ValueError(msg)

            self._load_model()

            # clustering the datasets
            if self.clustering_method is not None:
                if self.clustering_method in ("mcl", "louvain"):
                    _log.info("Loading clusters")
                    self._precluster(self.dataset_train)

                    if self.dataset_val is not None:
                        self._precluster(self.dataset_val)
                    else:
                        _log.warning("No validation dataset given. Randomly splitting training set in training set and validation set.")
                        self.dataset_train, self.dataset_val = _divide_dataset(self.dataset_train, splitsize=self.val_size)

                    if self.dataset_test is not None:
                        self._precluster(self.dataset_test)
                else:
                    msg = f"Invalid node clustering method: {self.clustering_method}. Please set clustering_method to 'mcl', 'louvain' or None."
                    raise ValueError(msg)

        else:
            if self.neuralnet is None:
                msg = "No neural network class found. Please add it to complete loading the pretrained model."
                raise ValueError(msg)
            if self.dataset_test is None:
                msg = "No dataset_test found. Please add it to evaluate the pretrained model."
                raise ValueError(msg)
            if self.dataset_train is not None:
                self.dataset_train = None
                _log.warning("Pretrained model loaded: dataset_train will be ignored.")
            if self.dataset_val is not None:
                self.dataset_val = None
                _log.warning("Pretrained model loaded: dataset_val will be ignored.")
            self._init_from_dataset(self.dataset_test)
            self._load_params()
            self._load_pretrained_model()

    def _init_output_exporters(self, output_exporters: list[OutputExporter] | None) -> None:
        if output_exporters is not None:
            self._output_exporters = OutputExporterCollection(*output_exporters)
        else:
            self._output_exporters = OutputExporterCollection(HDF5OutputExporter("./output"))

    def _init_datasets(
        self,
        dataset_train: GraphDataset | GridDataset,
        dataset_val: GraphDataset | GridDataset | None,
        dataset_test: GraphDataset | GridDataset | None,
        val_size: int | float | None,
        test_size: int | float | None,
    ) -> None:
        self._check_dataset_equivalence(dataset_train, dataset_val, dataset_test)

        self.dataset_train = dataset_train
        self.dataset_test = dataset_test
        self.dataset_val = dataset_val
        self.val_size = val_size
        self.test_size = test_size

        # Divide datasets where necessary.
        if test_size is not None:
            if dataset_test is None:
                self.dataset_train, self.dataset_test = _divide_dataset(dataset_train, test_size)
            else:
                _log.warning("Test dataset was provided to Trainer; test_size parameter is ignored.")

        if val_size is not None:
            if dataset_val is None:
                self.dataset_train, self.dataset_val = _divide_dataset(dataset_train, val_size)
            else:
                _log.warning("Validation dataset was provided to Trainer; val_size parameter is ignored.")

    def _init_from_dataset(self, dataset: GraphDataset | GridDataset) -> None:
        if isinstance(dataset, GraphDataset):
            self.clustering_method = dataset.clustering_method
            self.node_features = dataset.node_features
            self.edge_features = dataset.edge_features
            self.features = None
            self.features_transform = dataset.features_transform
            self.means = dataset.means
            self.devs = dataset.devs

        elif isinstance(dataset, GridDataset):
            self.clustering_method = None
            self.node_features = None
            self.edge_features = None
            self.features = dataset.features
            self.features_transform = None
            self.means = None
            self.devs = None
        else:
            msg = f"Incorrect `dataset` type provided: {type(dataset)}. Please provide a `GridDataset` or `GraphDataset` object instead."
            raise TypeError(msg)

        self.target = dataset.target
        self.target_transform = dataset.target_transform
        self.task = dataset.task
        self.classes = dataset.classes
        self.classes_to_index = dataset.classes_to_index

    def _load_model(self) -> None:
        """Loads the neural network model."""
        self._put_model_to_device(self.dataset_train)
        self.configure_optimizers()
        self.set_lossfunction()

    def _check_dataset_equivalence(
        self,
        dataset_train: GraphDataset | GridDataset,
        dataset_val: GraphDataset | GridDataset,
        dataset_test: GraphDataset | GridDataset,
    ) -> None:
        """Check dataset_train type and train_source parameter settings."""
        # dataset_train is None when pretrained_model is set
        if dataset_train is None:
            # only check the test dataset
            if dataset_test is None:
                msg = "Please provide at least a train or test dataset"
                raise ValueError(msg)
        else:
            # Make sure train dataset has valid type
            if not isinstance(dataset_train, GraphDataset) and not isinstance(dataset_train, GridDataset):
                msg = f"train dataset is not the right type {type(dataset_train)}. Make sure it's either GraphDataset or GridDataset"
                raise TypeError(msg)

            if dataset_val is not None:
                self._check_dataset_value(
                    dataset_train,
                    dataset_val,
                    type_dataset="valid",
                )

            if dataset_test is not None:
                self._check_dataset_value(
                    dataset_train,
                    dataset_test,
                    type_dataset="test",
                )

    def _check_dataset_value(
        self,
        dataset_train: GraphDataset | GridDataset,
        dataset_check: GraphDataset | GridDataset,
        type_dataset: str,
    ) -> None:
        """Check valid/test dataset settings."""
        # Check train_source parameter in valid/test is set.
        if dataset_check.train_source is None:
            msg = f"{type_dataset} dataset has train_source parameter set to None. Make sure to set it as a valid training data source."
            raise ValueError(msg)
        # Check train_source parameter in valid/test is equivalent to train which passed to Trainer.
        if dataset_check.train_source != dataset_train:
            msg = f"{type_dataset} dataset has different train_source parameter from Trainer. Make sure to assign equivalent train_source in Trainer."
            raise ValueError(msg)

    def _load_pretrained_model(self) -> None:
        """Loads pretrained model."""
        self.test_loader = DataLoader(self.dataset_test, pin_memory=self.cuda)
        _log.info("Testing set loaded\n")
        self._put_model_to_device(self.dataset_test)

        # load the model and the optimizer state
        self.optimizer = self.optimizer(
            self.model.parameters(),
            lr=self.lr,
            weight_decay=self.weight_decay,
        )
        self.optimizer.load_state_dict(self.opt_loaded_state_dict)
        self.model.load_state_dict(self.model_load_state_dict)

    def _precluster(self, dataset: GraphDataset) -> None:
        """Pre-clusters nodes of the graphs."""
        for fname, mol in tqdm(dataset.index_entries):
            data = dataset.load_one_graph(fname, mol)

            if data is None:
                f5 = h5py.File(fname, "a")
                try:
                    _log.info(f"deleting {mol}")
                    del f5[mol]
                except BaseException:  # noqa: BLE001
                    _log.info(f"{mol} not found")
                f5.close()
                continue

            f5 = h5py.File(fname, "a")
            grp = f5[mol]
            clust_grp = grp.require_group("clustering")

            if self.clustering_method.lower() in clust_grp:
                del clust_grp[self.clustering_method.lower()]

            method_grp = clust_grp.create_group(self.clustering_method.lower())
            cluster = community_detection(data.edge_index, data.num_nodes, method=self.clustering_method)
            method_grp.create_dataset("depth_0", data=cluster.cpu())
            data = community_pooling(cluster, data)
            cluster = community_detection(data.edge_index, data.num_nodes, method=self.clustering_method)
            method_grp.create_dataset("depth_1", data=cluster.cpu())

            f5.close()

    def _put_model_to_device(self, dataset: GraphDataset | GridDataset) -> None:
        """
        Puts the model on the available device.

        Args:
            dataset (:class:`GraphDataset` | :class:`GridDataset`): GraphDataset object.

        Raises:
            ValueError: Incorrect output shape
        """
        # regression mode
        if self.task == targets.REGRESS:
            self.output_shape = 1

        # classification mode
        elif self.task == targets.CLASSIF:
            self.output_shape = len(self.classes)

        # the target values are optional
        if dataset.get(0).y is not None:
            target_shape = dataset.get(0).y.shape[0]
        else:
            target_shape = None

        if isinstance(dataset, GraphDataset):
            num_node_features = dataset.get(0).num_features
            num_edge_features = len(dataset.edge_features)

            self.model = self.neuralnet(num_node_features, self.output_shape, num_edge_features).to(self.device)

        elif isinstance(dataset, GridDataset):
            _, num_features, box_width, box_height, box_depth = dataset.get(0).x.shape

            self.model = self.neuralnet(num_features, (box_width, box_height, box_depth)).to(self.device)
        else:
            raise TypeError(type(dataset))

        # multi-gpu
        if self.ngpu > 1:
            ids = list(range(self.ngpu))
            self.model = nn.DataParallel(self.model, device_ids=ids).to(self.device)

        # check for compatibility
        for output_exporter in self._output_exporters:
            if not output_exporter.is_compatible_with(self.output_shape, target_shape):
                msg = (
                    f"Output exporter of type {type(output_exporter)}\n\t"
                    f"is not compatible with output shape {self.output_shape}\n\t"
                    f"and target shape {target_shape}."
                )
                raise ValueError(msg)

    def configure_optimizers(
        self,
        optimizer: torch.optim = None,
        lr: float = 0.001,
        weight_decay: float = 1e-05,
    ) -> None:
        """
        Configure optimizer and its main parameters.

        Args:
            optimizer (:class:`torch.optim`, optional): PyTorch optimizer object. If none, defaults to :class:`torch.optim.Adam`.
                Defaults to None.

            lr (float, optional): Learning rate. Defaults to 0.001.

            weight_decay (float, optional): Weight decay (L2 penalty).
                Weight decay is fundamental for GNNs, otherwise, parameters can become too big and the gradient may explode. Defaults to 1e-05.
        """
        self.lr = lr
        self.weight_decay = weight_decay

        if optimizer is None:
            self.optimizer = torch.optim.Adam(self.model.parameters(), lr=lr, weight_decay=weight_decay)
        else:
            try:
                self.optimizer = optimizer(self.model.parameters(), lr=lr, weight_decay=weight_decay)
            except Exception as e:
                _log.error(e)
                _log.info("Invalid optimizer. Please use only optimizers classes from torch.optim package.")
                raise

    def set_lossfunction(  # noqa: C901
        self,
        lossfunction: nn.modules.loss._Loss | None = None,
        override_invalid: bool = False,
    ) -> None:
        """
        Set the loss function.

        Args:
            lossfunction (optional): Make sure to use a loss function that is appropriate for
                your task (classification or regression). All loss functions
                from torch.nn.modules.loss are listed as belonging to either
                category (or to neither) and an exception is raised if an invalid
                loss function is chosen for the set task.
                Default for regression: MSELoss.
                Default for classification: CrossEntropyLoss.
                Defaults to None.
            override_invalid (bool, optional): If True, loss functions that are considered
                invalid for the task do no longer automaticallt raise an exception.
                Defaults to False.
        """
        default_regression_loss = nn.MSELoss
        default_classification_loss = nn.CrossEntropyLoss

        def _invalid_loss() -> None:
            if override_invalid:
                _log.warning(
                    f"The provided loss function ({lossfunction}) is not appropriate for {self.task} tasks.\n\t"
                    "You have set override_invalid to True, so the training will run with this loss function nonetheless.\n\t"
                    "This will likely cause other errors or exceptions down the line.",
                )
            else:
                invalid_loss_error = (
                    f"The provided loss function ({lossfunction}) is not appropriate for {self.task} tasks.\n\t"
                    "If you want to use this loss function anyway, set override_invalid to True."
                )
                _log.error(invalid_loss_error)
                raise ValueError(invalid_loss_error)

        # check for custom/invalid loss functions
        if lossfunction in losses.other_losses:
            _invalid_loss()
        elif lossfunction not in (losses.regression_losses + losses.classification_losses):
            custom_loss = True
        else:
            custom_loss = False

        # set regression loss
        if self.task == targets.REGRESS:
            if lossfunction is None:
                lossfunction = default_regression_loss
                _log.info(f"No loss function provided, the default loss function for {self.task} tasks is used: {lossfunction}")
            elif custom_loss:
                custom_loss_warning = (
                    f"The provided loss function ({lossfunction}) is not part of the default list.\n\t"
                    f"Please ensure that this loss function is appropriate for {self.task} tasks.\n\t"
                )
                _log.warning(custom_loss_warning)
            elif lossfunction not in losses.regression_losses:
                _invalid_loss()
            self.lossfunction = lossfunction()

        # Set classification loss
        elif self.task == targets.CLASSIF:
            if lossfunction is None:
                lossfunction = default_classification_loss
                _log.info(f"No loss function provided, the default loss function for {self.task} tasks is used: {lossfunction}")
            elif custom_loss:
                custom_loss_warning = (
                    f"The provided loss function ({lossfunction}) is not part of the default list.\n\t"
                    f"Please ensure that this loss function is appropriate for {self.task} tasks.\n\t"
                )
                _log.warning(custom_loss_warning)
            elif lossfunction not in losses.classification_losses:
                _invalid_loss()

            if not self.class_weights:
                self.lossfunction = lossfunction()
            else:
                self.lossfunction = lossfunction  # weights will be set in the train() method

    def train(  # noqa: PLR0915, C901
        self,
        nepoch: int = 1,
        batch_size: int = 32,
        shuffle: bool = True,
        earlystop_patience: int | None = None,
        earlystop_maxgap: float | None = None,
        min_epoch: int = 10,
        validate: bool = False,
        num_workers: int = 0,
        best_model: bool = True,
        filename: str | None = "model.pth.tar",
    ) -> None:
        """
        Performs the training of the model.

        Args:
            nepoch (int, optional): Maximum number of epochs to run.
                        Defaults to 1.
            batch_size (int, optional): Sets the size of the batch.
                        Defaults to 32.
            shuffle (bool, optional): Whether to shuffle the training dataloaders data (train set and validation set).
                        Default: True.
            earlystop_patience (int | None, optional): Training ends if the model has run for this number of epochs without improving the validation loss.
                        Defaults to None.
            earlystop_maxgap (float | None, optional): Training ends if the difference between validation and training loss exceeds this value.
                        Defaults to None.
            min_epoch (float, optional): Minimum epoch to be reached before looking at maxgap.
                        Defaults to 10.
            validate (bool, optional): Perform validation on independent data set (requires a validation data set).
                        Defaults to False.
            num_workers (int, optional): How many subprocesses to use for data loading. 0 means that the data will be loaded in the main process.
                        Defaults to 0.
            best_model (bool, optional):
                        If True, the best model (in terms of validation loss) is selected for later testing or saving.
                        If False, the last model tried is selected.
                        Defaults to True.
            filename (str, optional): Name of the file where to save the selected model. If not None, the model is saved to `filename`.
                If None, the model is not saved. Defaults to 'model.pth.tar'.
        """
        if self.dataset_train is None:
            msg = "No training dataset provided."
            raise ValueError(msg)

        self.data_type = type(self.dataset_train)
        self.batch_size_train = batch_size
        self.shuffle = shuffle

        self.train_loader = DataLoader(
            self.dataset_train,
            batch_size=self.batch_size_train,
            shuffle=self.shuffle,
            num_workers=num_workers,
            pin_memory=self.cuda,
        )
        _log.info("Training set loaded\n")

        if self.dataset_val is not None:
            self.valid_loader = DataLoader(
                self.dataset_val,
                batch_size=self.batch_size_train,
                shuffle=self.shuffle,
                num_workers=num_workers,
                pin_memory=self.cuda,
            )
            _log.info("Validation set loaded\n")
        else:
            self.valid_loader = None
            _log.info("No validation set provided\n")
            _log.warning(
                "Training data will be used both for learning and model selection, which may lead to overfitting.\n"
                "It is usually preferable to use a validation set during the training phase.",
            )

        # Assign weights to each class
        if self.task == targets.CLASSIF and self.class_weights:
            targets_all = [batch.y for batch in self.train_loader]

            targets_all = torch.cat(targets_all).squeeze().tolist()
            self.weights = torch.tensor([targets_all.count(i) for i in self.classes], dtype=torch.float32)
            _log.info(f"class occurences: {self.weights}")
            self.weights = 1.0 / self.weights
            self.weights = self.weights / self.weights.sum()
            _log.info(f"class weights: {self.weights}")

            try:
                self.lossfunction = self.lossfunction(weight=self.weights.to(self.device))  # Check whether loss allows for weighted classes
            except TypeError as e:
                weight_error = (
                    f"Loss function {self.lossfunction} does not allow for weighted classes.\n\t"
                    "Please use a different loss function or set class_weights to False.\n"
                )
                _log.error(weight_error)
                raise ValueError(weight_error) from e
        else:
            self.weights = None

        train_losses = []
        valid_losses = []
        saved_model = False

        if earlystop_patience or earlystop_maxgap:
            early_stopping = EarlyStopping(
                patience=earlystop_patience,
                maxgap=earlystop_maxgap,
                min_epoch=min_epoch,
                trace_func=_log.info,
            )
        else:
            early_stopping = None

        with self._output_exporters:
            # Number of epochs
            self.nepoch = nepoch
            _log.info("Epoch 0:")
            self._eval(self.train_loader, 0, "training")
            if validate:
                if self.valid_loader is None:
                    msg = "No validation dataset provided."
                    raise ValueError(msg)
                self._eval(self.valid_loader, 0, "validation")

            # Loop over epochs
            for epoch in range(1, nepoch + 1):
                _log.info(f"Epoch {epoch}:")

                # Set the module in training mode
                self.model.train()
                loss_ = self._epoch(epoch, "training")
                train_losses.append(loss_)

                # Validate the model
                if validate:
                    loss_ = self._eval(self.valid_loader, epoch, "validation")
                    valid_losses.append(loss_)
                    if best_model and min(valid_losses) == loss_:
                        checkpoint_model = self._save_model()
                        saved_model = True
                        self.epoch_saved_model = epoch
                        _log.info(f"Best model saved at epoch # {self.epoch_saved_model}.")
                    # check early stopping criteria (in validation case only)
                    if early_stopping:
                        # compare last validation and training loss
                        early_stopping(epoch, valid_losses[-1], train_losses[-1])
                        if early_stopping.early_stop:
                            break

                elif best_model:  # if no validation set, save the best performing model on the training set
                    if min(train_losses) == loss_:
                        checkpoint_model = self._save_model()
                        saved_model = True
                        self.epoch_saved_model = epoch
                        _log.info(f"Best model saved at epoch # {self.epoch_saved_model}.")

            # Save the last model
            if best_model is False or not saved_model:
                checkpoint_model = self._save_model()
                self.epoch_saved_model = epoch
                _log.info(f"Last model saved at epoch # {self.epoch_saved_model}.")
                if not saved_model:
                    warnings.warn(
                        "A model has been saved but the validation and/or the training losses were NaN;\n\t"
                        "try to increase the cutoff distance during the data processing or the number of data points during the training.",
                    )

        # Now that the training loop is over, save the model
        if filename:
            torch.save(checkpoint_model, filename)
        self.opt_loaded_state_dict = checkpoint_model["optimizer_state"]
        self.model_load_state_dict = checkpoint_model["model_state"]
        self.optimizer.load_state_dict(self.opt_loaded_state_dict)
        self.model.load_state_dict(self.model_load_state_dict)

    def _epoch(self, epoch_number: int, pass_name: str) -> float | None:
        """
        Runs a single epoch.

        Args:
            epoch_number (int): the current epoch number
            pass_name (str): 'training', 'validation' or 'testing'

        Returns:
            Running loss.
        """
        sum_of_losses = 0
        count_predictions = 0
        target_vals = []
        outputs = []
        entry_names = []
        t0 = time()
        for data_batch in self.train_loader:
            if self.cuda:
                data_batch = data_batch.to(self.device, non_blocking=True)  # noqa: PLW2901
            self.optimizer.zero_grad()
            pred = self.model(data_batch)
            pred, data_batch.y = self._format_output(pred, data_batch.y)
            loss_ = self.lossfunction(pred, data_batch.y)
            loss_.backward()
            self.optimizer.step()
            count_predictions += pred.shape[0]

            # convert mean back to sum
            sum_of_losses += loss_.detach().item() * pred.detach().shape[0]
            target_vals += data_batch.y.detach().cpu().numpy().tolist()

            # Get the outputs for export
            # Remember that non-linear activation is automatically applied in CrossEntropyLoss
            if self.task == targets.CLASSIF:
                pred = F.softmax(pred.detach(), dim=1)
            else:
                pred = pred.detach().reshape(-1)
            outputs += pred.cpu().numpy().tolist()

            # Get the name
            entry_names += data_batch.entry_names

        dt = time() - t0
        if count_predictions > 0:
            epoch_loss = sum_of_losses / count_predictions
        else:
            epoch_loss = None

        self._output_exporters.process(
            pass_name,
            epoch_number,
            entry_names,
            outputs,
            target_vals,
            epoch_loss,
        )
        self._log_epoch_data(pass_name, epoch_loss, dt)

        return epoch_loss

    def _eval(
        self,
        loader: DataLoader,
        epoch_number: int,
        pass_name: str,
    ) -> float | None:
        """
        Evaluates the model.

        Args:
            loader (Dataloader): Data to evaluate on.
            epoch_number (int): Number for this epoch, used for storing the outputs.
            pass_name (str): 'training', 'validation' or 'testing'

        Returns:
            Running loss.
        """
        # Sets the module in evaluation mode
        self.model.eval()
        loss_func = self.lossfunction
        target_vals = []
        outputs = []
        entry_names = []
        sum_of_losses = 0
        count_predictions = 0
        t0 = time()
        for data_batch in loader:
            if self.cuda:
                data_batch = data_batch.to(self.device, non_blocking=True)  # noqa: PLW2901
            pred = self.model(data_batch)
            pred, y = self._format_output(pred, data_batch.y)

            # Check if a target value was provided (i.e. benchmarck scenario)
            if y is not None:
                target_vals += y.cpu().numpy().tolist()
                loss_ = loss_func(pred, y)
                count_predictions += pred.shape[0]
                sum_of_losses += loss_.detach().item() * pred.shape[0]
            else:
                target_vals += [None] * pred.shape[0]
                eval_loss = None

            # Get the outputs for export
            # Remember that non-linear activation is automatically applied in CrossEntropyLoss
            if self.task == targets.CLASSIF:
                pred = F.softmax(pred.detach(), dim=1)
            else:
                pred = pred.detach().reshape(-1)
            outputs += pred.cpu().numpy().tolist()

            # get the name
            entry_names += data_batch.entry_names

        dt = time() - t0
        if count_predictions > 0:
            eval_loss = sum_of_losses / count_predictions
        else:
            eval_loss = None

        self._output_exporters.process(
            pass_name,
            epoch_number,
            entry_names,
            outputs,
            target_vals,
            eval_loss,
        )
        self._log_epoch_data(pass_name, eval_loss, dt)

        return eval_loss

    @staticmethod
    def _log_epoch_data(stage: str, loss: float, time: float) -> None:
        """
        Prints the data of each epoch.

        Args:
            stage (str): Train or valid.
            loss (float): Loss during that epoch.
            time (float): Timing of the epoch.
        """
        _log.info(f"{stage} loss {loss} | time {time}")

    def _format_output(self, pred, target=None):  # noqa: ANN001, ANN202
        """Format the network output depending on the task (classification/regression)."""
        if (self.task == targets.CLASSIF) and (target is not None):
            # For categorical cross entropy, the target must be a one-dimensional tensor
            # of class indices with type long and the output should have raw, unnormalized values
            target = torch.tensor([self.classes_to_index[x] if isinstance(x, str) else self.classes_to_index[int(x)] for x in target])
            if isinstance(self.lossfunction, nn.BCELoss | nn.BCEWithLogitsLoss):
                # # pred must be in (0,1) range and target must be float with same shape as pred
                msg = (
                    "BCELoss and BCEWithLogitsLoss are currently not supported.\n\t"
                    "For further details see: https://github.com/DeepRank/deeprank2/issues/318"
                )
                raise ValueError(msg)

            if isinstance(self.lossfunction, losses.classification_losses) and not isinstance(self.lossfunction, losses.classification_tested):
                msg = (
                    f"{self.lossfunction} is currently not supported.\n\t"
                    f"Supported loss functions for classification: {losses.classification_tested}.\n\t"
                    "Implementation of other loss functions requires adaptation of Trainer._format_output."
                )
                raise ValueError(msg)

        elif self.task == targets.REGRESS:
            pred = pred.reshape(-1)

        if target is not None:
            target = target.to(self.device)

        return pred, target

    def test(
        self,
        batch_size: int = 32,
        num_workers: int = 0,
    ) -> None:
        """
        Performs the testing of the model.

        Args:
            batch_size (int, optional): Sets the size of the batch.
                        Defaults to 32.
            num_workers (int, optional): How many subprocesses to use for data loading. 0 means that the data will be loaded in the main process.
                        Defaults to 0.
        """
        if (not self.pretrained_model) and (not self.model_load_state_dict):
            msg = "No pretrained model provided and no training performed. Please provide a pretrained model or train the model before testing."
            raise ValueError(msg)

        self.batch_size_test = batch_size

        if self.dataset_test is not None:
            _log.info("Loading independent testing dataset...")

            self.test_loader = DataLoader(
                self.dataset_test,
                batch_size=self.batch_size_test,
                num_workers=num_workers,
                pin_memory=self.cuda,
            )
            _log.info("Testing set loaded\n")
        else:
            _log.error("No test dataset provided.")
            msg = "No test dataset provided."
            raise ValueError(msg)

        with self._output_exporters:
            # Run test
            self._eval(self.test_loader, self.epoch_saved_model, "testing")

    def _load_params(self) -> None:
        """Loads the parameters of a pretrained model."""
        if torch.cuda.is_available():
            state = torch.load(self.pretrained_model)
        else:
            state = torch.load(self.pretrained_model, map_location=torch.device("cpu"))

        self.data_type = state["data_type"]
        self.model_load_state_dict = state["model_state"]
        self.optimizer = type(state["optimizer"])
        self.opt_loaded_state_dict = state["optimizer_state"]
        self.lossfunction = state["lossfunction"]
        self.target = state["target"]
        self.target_transform = state["target_transform"]
        self.task = state["task"]
        self.classes = state["classes"]
        self.classes_to_index = state["classes_to_index"]
        self.class_weights = state["class_weights"]
        self.batch_size_train = state["batch_size_train"]
        self.batch_size_test = state["batch_size_test"]
        self.val_size = state["val_size"]
        self.test_size = state["test_size"]
        self.lr = state["lr"]
        self.weight_decay = state["weight_decay"]
        self.epoch_saved_model = state["epoch_saved_model"]
        self.subset = state["subset"]
        self.shuffle = state["shuffle"]
        self.clustering_method = state["clustering_method"]
        self.node_features = state["node_features"]
        self.edge_features = state["edge_features"]
        self.features = state["features"]
        self.features_transform = state["features_transform"]
        self.means = state["means"]
        self.devs = state["devs"]
        self.cuda = state["cuda"]
        self.ngpu = state["ngpu"]

    def _save_model(self) -> dict[str, Any]:
        """
        Saves the model to a file.

        Args:
            filename (str, optional): Name of the file. Defaults to None.
        """
        features_transform_to_save = copy.deepcopy(self.features_transform)
        # prepare transform dictionary for being saved
        if features_transform_to_save:
            for key in features_transform_to_save.values():
                if key["transform"] is None:
                    continue
                # Serialize the function
                serialized_func = dill.dumps(key["transform"])
                # Deserialize the function
                deserialized_func = dill.loads(serialized_func)  # noqa: S301
                str_expr = inspect.getsource(deserialized_func)
                match = re.search(r"[\"|\']transform[\"|\']:.*(lambda.*).*,.*[\"|\']standardize[\"|\'].*", str_expr).group(1)
                key["transform"] = match

        state = {
            "data_type": self.data_type,
            "model_state": self.model.state_dict(),
            "optimizer": self.optimizer,
            "optimizer_state": self.optimizer.state_dict(),
            "lossfunction": self.lossfunction,
            "target": self.target,
            "target_transform": self.target_transform,
            "task": self.task,
            "classes": self.classes,
            "classes_to_index": self.classes_to_index,
            "class_weights": self.class_weights,
            "batch_size_train": self.batch_size_train,
            "batch_size_test": self.batch_size_test,
            "val_size": self.val_size,
            "test_size": self.test_size,
            "lr": self.lr,
            "weight_decay": self.weight_decay,
            "epoch_saved_model": self.epoch_saved_model,
            "subset": self.subset,
            "shuffle": self.shuffle,
            "clustering_method": self.clustering_method,
            "node_features": self.node_features,
            "edge_features": self.edge_features,
            "features": self.features,
            "features_transform": features_transform_to_save,
            "means": self.means,
            "devs": self.devs,
            "cuda": self.cuda,
            "ngpu": self.ngpu,
        }

        return state  # noqa:RET504 (unnecessary-assign)


def _divide_dataset(
    dataset: GraphDataset | GridDataset,
    splitsize: float | int | None = None,
) -> tuple[GraphDataset, GraphDataset] | tuple[GridDataset, GridDataset]:
    """Divides the dataset into a training set and an evaluation set.

    Args:
        dataset (:class:`GraphDataset` | :class:`GridDataset`): Input dataset to be split into training and validation data.
        splitsize (float | int | None, optional): Fraction of dataset (if float) or number of datapoints (if int) to use for validation.
            Defaults to None.
    """
    if splitsize is None:
        splitsize = 0.25
    full_size = len(dataset)

    # find number of datapoints to include in training dataset
    if isinstance(splitsize, float):
        n_split = int(splitsize * full_size)
    elif isinstance(splitsize, int):
        n_split = splitsize
    else:
        msg = f"type(splitsize) must be float, int or None ({type(splitsize)} detected.)"
        raise TypeError(msg)

    # raise exception if no training data or negative validation size
    if n_split >= full_size or n_split < 0:
        msg = f"Invalid Split size: {n_split}.\n"
        f"Split size must be a float between 0 and 1 OR an int smaller than the size of the dataset ({full_size} datapoints)"
        raise ValueError(msg)

    if splitsize == 0:  # i.e. the fraction of splitsize was so small that it rounded to <1 datapoint
        dataset_main = dataset
        dataset_split = None
    else:
        indices = np.arange(full_size)
        np.random.default_rng().shuffle(indices)

        dataset_main = copy.deepcopy(dataset)
        dataset_main.index_entries = [dataset.index_entries[i] for i in indices[n_split:]]

        dataset_split = copy.deepcopy(dataset)
        dataset_split.index_entries = [dataset.index_entries[i] for i in indices[:n_split]]

    return dataset_main, dataset_split



import re
from enum import Enum
from typing import Any

# ruff: noqa: D101


class PatchActionType(Enum):
    MODIFY = 1
    ADD = 2


class PatchSelection:
    def __init__(self, residue_type: str, atom_name: str):
        self.residue_type = residue_type
        self.atom_name = atom_name


class PatchAction:
    def __init__(self, type_: str, selection: PatchSelection, kwargs: dict[str, Any]):
        self.type = type_
        self.selection = selection
        self.kwargs = kwargs

    def __contains__(self, key: str):
        return key in self.kwargs

    def __getitem__(self, key: str):
        return self.kwargs[key]


class PatchParser:
    STRING_VAR_PATTERN = re.compile(r"([A-Z]+)=([A-Z0-9]+)")
    NUMBER_VAR_PATTERN = re.compile(r"([A-Z]+)=(\-?[0-9]+\.[0-9]+)")
    ACTION_PATTERN = re.compile(r"^([A-Z]{3,4})\s+([A-Z]+)\s+ATOM\s+([A-Z0-9]{1,3})\s+(.*)$")

    @staticmethod
    def _parse_action_type(s: str) -> PatchActionType:
        for type_ in PatchActionType:
            if type_.name == s:
                return type_

        msg = f"Unmatched residue action: {s!r}"
        raise ValueError(msg)

    @staticmethod
    def parse(file_: str) -> list[PatchAction]:
        result = []
        for line in file_:
            if line.startswith(("#", "!")) or len(line.strip()) == 0:
                continue

            m = PatchParser.ACTION_PATTERN.match(line)
            if not m:
                msg = f"Unmatched patch action: {line!r}"
                raise ValueError(msg)

            residue_type = m.group(1)
            action_type = PatchParser._parse_action_type(m.group(2))
            atom_name = m.group(3)

            kwargs = {}
            for w in PatchParser.STRING_VAR_PATTERN.finditer(m.group(4)):
                kwargs[w.group(1)] = w.group(2)
            for w in PatchParser.NUMBER_VAR_PATTERN.finditer(m.group(4)):
                kwargs[w.group(1)] = float(w.group(2))

            result.append(PatchAction(action_type, PatchSelection(residue_type, atom_name), kwargs))
        return result



import torch
from torch import nn

# ruff: noqa: ANN001, ANN201

__author__ = "Daniel-Tobias Rademaker"


class GNNLayer(nn.Module):  # noqa: D101
    def __init__(
        self,
        nmb_edge_projection,
        nmb_hidden_attr,
        nmb_output_features,
        message_vector_length,
        nmb_mlp_neurons,
        act_fn=nn.SiLU(),  # noqa: B008
        is_last_layer=True,
    ):
        super().__init__()
        # The MLP that takes in atom-pairs and creates the Mij's
        self.edge_mlp = nn.Sequential(
            nn.Linear(nmb_edge_projection + nmb_hidden_attr * 2, nmb_mlp_neurons),
            act_fn,
            nn.Linear(nmb_mlp_neurons, message_vector_length),
            act_fn,
        )

        # The node-MLP, creates a new node-representation given the Mi's
        self.node_mlp = nn.Sequential(
            nn.BatchNorm1d(message_vector_length + nmb_hidden_attr),
            nn.Linear(message_vector_length + nmb_hidden_attr, nmb_mlp_neurons),
            act_fn,
            nn.Linear(nmb_mlp_neurons, nmb_mlp_neurons),
            act_fn,
            nn.Linear(nmb_mlp_neurons, nmb_hidden_attr),
        )

        # Only last layer have attention and output modules
        if is_last_layer:
            # attention mlp, to weight the ouput significance
            self.attention_mlp = nn.Sequential(
                nn.Linear(nmb_hidden_attr, nmb_mlp_neurons),
                act_fn,
                nn.Linear(nmb_mlp_neurons, 1),
                nn.Sigmoid(),
            )

            # Create the output vector per node we are interested in
            self.output_mlp = nn.Sequential(
                nn.Linear(nmb_hidden_attr, nmb_mlp_neurons),
                act_fn,
                nn.Linear(nmb_mlp_neurons, nmb_output_features),
            )

    # MLP that takes in the node-attributes of nodes (source + target), the edge attributes
    # and node attributes in order to create a 'message vector'between those
    # nodes
    def edge_model(self, edge_attr, hidden_features_source, hidden_features_target):
        cat = torch.cat([edge_attr, hidden_features_source, hidden_features_target], dim=1)
        return self.edge_mlp(cat)

    # A function that updates the node-attributes. Assumed that submessages
    # are already summed
    def node_model(self, summed_edge_message, hidden_features):
        cat = torch.cat([summed_edge_message, hidden_features], dim=1)
        output = self.node_mlp(cat)
        return hidden_features + output

    # Sums the individual sub-messages (multiple per node) into singel message
    # vector per node
    def sum_messages(self, edges, messages, nmb_nodes):
        row, _ = edges
        summed_messages_shape = (nmb_nodes, messages.size(1))
        result = messages.new_full(summed_messages_shape, 0)
        row = row.unsqueeze(-1).expand(-1, messages.size(1))
        result.scatter_add_(0, row, messages)
        return result

    # Runs the GNN
    # steps is number of times it exanges info with neighbors
    def update_nodes(self, edges, edge_attr, hidden_features, steps=1):
        (
            row,
            col,
        ) = edges  # a single edge is defined as the index of atom1 and the index of atom2
        h = hidden_features  # shortening the variable name
        # It is possible to run input through the same same layer multiple
        # times
        for _ in range(steps):
            node_pair_messages = self.edge_model(edge_attr, h[row], h[col])  # get all atom-pair messages
            # sum all messages per node to single message vector
            messages = self.sum_messages(edges, node_pair_messages, len(h))
            # Use the messages to update the node-attributes
            h = self.node_model(messages, h)
        return h

    # output, every node creates a prediction + an estimate how sure it is of
    # its prediction. Only done by last 'GNN layer'
    def output(self, hidden_features, get_attention=True):
        output = self.output_mlp(hidden_features)
        if get_attention:
            return output, self.attention_mlp(hidden_features)
        return output


class SuperGNN(nn.Module):  # noqa: D101
    def __init__(
        self,
        nmb_edge_attr,
        nmb_node_attr,
        nmb_hidden_attr,
        nmb_mlp_neurons,
        nmb_edge_projection,
        nmb_gnn_layers,
        nmb_output_features,
        message_vector_length,
        act_fn=nn.SiLU(),  # noqa: B008
    ):
        super().__init__()

        # Since edge_atributes go into every layer, it might be betetr to learn
        # a better/smarter representation of them first
        self.preproc_edge_mlp = nn.Sequential(
            nn.BatchNorm1d(nmb_edge_attr),
            nn.Linear(nmb_edge_attr, nmb_mlp_neurons),
            nn.BatchNorm1d(nmb_mlp_neurons),
            act_fn,
            nn.Linear(nmb_mlp_neurons, nmb_edge_projection),
            act_fn,
        )

        # Project the node_attributes to the same size as the hidden vector
        self.preproc_node_mlp = nn.Sequential(
            nn.BatchNorm1d(nmb_node_attr),
            nn.Linear(nmb_node_attr, nmb_mlp_neurons),
            nn.BatchNorm1d(nmb_mlp_neurons),
            act_fn,
            nn.Linear(nmb_mlp_neurons, nmb_hidden_attr),
            act_fn,
        )

        self.modlist = nn.ModuleList(
            [
                GNNLayer(
                    nmb_edge_projection,
                    nmb_hidden_attr,
                    nmb_output_features,
                    message_vector_length,
                    nmb_mlp_neurons,
                    is_last_layer=(gnn_layer == (nmb_gnn_layers - 1)),
                )
                for gnn_layer in range(nmb_gnn_layers)
            ],
        )

    # always use this function before running the GNN layers
    def preprocess(self, edge_attr, node_attr):
        edge_attr = self.preproc_edge_mlp(edge_attr)
        hidden_features = self.preproc_node_mlp(node_attr)
        return edge_attr, hidden_features

    # Runs data through layers and return output. Potentially, attention can
    # also be returned
    def run_through_network(self, edges, edge_attr, node_attr, with_output_attention=False):
        edge_attr, node_attr = self.preprocess(edge_attr, node_attr)
        for layer in self.modlist:
            node_attr = layer.update_nodes(edges, edge_attr, node_attr)
        if with_output_attention:
            representations, attention = self.modlist[-1].output(node_attr, True)  # (boolean-positional-value-in-call)
            return representations, attention
        return self.modlist[-1].output(node_attr, True)  # (boolean-positional-value-in-call)


class AlignmentGNN(SuperGNN):  # noqa: D101
    def __init__(
        self,
        nmb_edge_attr,
        nmb_node_attr,
        nmb_output_features,
        nmb_hidden_attr,
        message_vector_length,
        nmb_mlp_neurons,
        nmb_gnn_layers,
        nmb_edge_projection,
        act_fn=nn.SiLU(),  # noqa: B008
    ):
        super().__init__(
            nmb_edge_attr,
            nmb_node_attr,
            nmb_hidden_attr,
            nmb_mlp_neurons,
            nmb_edge_projection,
            nmb_gnn_layers,
            nmb_output_features,
            message_vector_length,
            act_fn,
        )

    # Run over all layers, and return the ouput vectors
    def forward(self, edges, edge_attr, node_attr):
        return self.run_through_network(edges, edge_attr, node_attr)



import logging

import freesasa
import numpy as np

from deeprank2.domain import nodestorage as Nfeat
from deeprank2.molstruct.atom import Atom
from deeprank2.molstruct.residue import Residue, SingleResidueVariant
from deeprank2.utils.graph import Graph

freesasa.setVerbosity(freesasa.nowarnings)
logging.getLogger(__name__)


def add_sasa(pdb_path: str, graph: Graph) -> None:  # noqa:D103
    structure = freesasa.Structure(pdb_path)
    result = freesasa.calc(structure)

    for node in graph.nodes:
        if isinstance(node.id, Residue):
            residue = node.id
            selection = (f"residue, (resi {residue.number_string}) and (chain {residue.chain.id})",)
            area = freesasa.selectArea(selection, structure, result)["residue"]

        elif isinstance(node.id, Atom):
            atom = node.id
            residue = atom.residue
            selection = (f"atom, (name {atom.name}) and (resi {residue.number_string}) and (chain {residue.chain.id})",)
            area = freesasa.selectArea(selection, structure, result)["atom"]

        else:
            msg = f"Unexpected node type: {type(node.id)}"
            raise TypeError(msg)

        if np.isnan(area):
            msg = f"freesasa returned {area} for {residue}"
            raise ValueError(msg)
        node.features[Nfeat.SASA] = area


def add_bsa(graph: Graph) -> None:  # noqa:D103
    sasa_complete_structure = freesasa.Structure()
    sasa_chain_structures = {}

    for node in graph.nodes:
        if isinstance(node.id, Residue):
            residue = node.id
            chain_id = residue.chain.id
            if chain_id not in sasa_chain_structures:
                sasa_chain_structures[chain_id] = freesasa.Structure()

            for atom in residue.atoms:
                sasa_chain_structures[chain_id].addAtom(
                    atom.name,
                    atom.residue.amino_acid.three_letter_code,
                    atom.residue.number,
                    atom.residue.chain.id,
                    atom.position[0],
                    atom.position[1],
                    atom.position[2],
                )
                sasa_complete_structure.addAtom(
                    atom.name,
                    atom.residue.amino_acid.three_letter_code,
                    atom.residue.number,
                    atom.residue.chain.id,
                    atom.position[0],
                    atom.position[1],
                    atom.position[2],
                )

        elif isinstance(node.id, Atom):
            atom = node.id
            residue = atom.residue
            chain_id = residue.chain.id
            if chain_id not in sasa_chain_structures:
                sasa_chain_structures[chain_id] = freesasa.Structure()

            sasa_chain_structures[chain_id].addAtom(
                atom.name,
                atom.residue.amino_acid.three_letter_code,
                atom.residue.number,
                atom.residue.chain.id,
                atom.position[0],
                atom.position[1],
                atom.position[2],
            )
            sasa_complete_structure.addAtom(
                atom.name,
                atom.residue.amino_acid.three_letter_code,
                atom.residue.number,
                atom.residue.chain.id,
                atom.position[0],
                atom.position[1],
                atom.position[2],
            )

            area_key = "atom"
            selection = f"atom, (name {atom.name}) and (resi {atom.residue.number_string}) and (chain {atom.residue.chain.id})"
        else:
            msg = f"Unexpected node type: {type(node.id)}"
            raise TypeError(msg)

    sasa_complete_result = freesasa.calc(sasa_complete_structure)
    sasa_chain_results = {chain_id: freesasa.calc(structure) for chain_id, structure in sasa_chain_structures.items()}

    for node in graph.nodes:
        if isinstance(node.id, Residue):
            residue = node.id
            chain_id = residue.chain.id
            area_key = "residue"
            selection = (f"residue, (resi {residue.number_string}) and (chain {residue.chain.id})",)

        elif isinstance(node.id, Atom):
            atom = node.id
            chain_id = atom.residue.chain.id
            area_key = "atom"
            selection = (f"atom, (name {atom.name}) and (resi {atom.residue.number_string}) and (chain {atom.residue.chain.id})",)

        area_monomer = freesasa.selectArea(selection, sasa_chain_structures[chain_id], sasa_chain_results[chain_id])[area_key]
        area_multimer = freesasa.selectArea(selection, sasa_complete_structure, sasa_complete_result)[area_key]

        node.features[Nfeat.BSA] = area_monomer - area_multimer


def add_features(
    pdb_path: str,
    graph: Graph,
    single_amino_acid_variant: SingleResidueVariant | None = None,  # noqa: ARG001
) -> None:
    """Calculates the Buried Surface Area (BSA) and the Solvent Accessible Surface Area (SASA)."""
    # BSA
    add_bsa(graph)

    # SASA
    add_sasa(pdb_path, graph)



import logging
import os

from deeprank2.molstruct.atom import Atom
from deeprank2.molstruct.residue import Residue
from deeprank2.utils.parsing.patch import PatchActionType, PatchParser
from deeprank2.utils.parsing.residue import ResidueClassParser
from deeprank2.utils.parsing.top import TopParser
from deeprank2.utils.parsing.vdwparam import ParamParser, VanderwaalsParam

_log = logging.getLogger(__name__)

_forcefield_directory_path = os.path.realpath(os.path.join(os.path.dirname(__file__), "../../domain/forcefield"))


class AtomicForcefield:  # noqa: D101
    def __init__(self):
        top_path = os.path.join(_forcefield_directory_path, "protein-allhdg5-5_new.top")
        with open(top_path, encoding="utf-8") as f:
            self._top_rows = {(row.residue_name, row.atom_name): row for row in TopParser.parse(f)}

        patch_path = os.path.join(_forcefield_directory_path, "patch.top")
        with open(patch_path, encoding="utf-8") as f:
            self._patch_actions = PatchParser.parse(f)

        residue_class_path = os.path.join(_forcefield_directory_path, "residue-classes")
        with open(residue_class_path, encoding="utf-8") as f:
            self._residue_class_criteria = ResidueClassParser.parse(f)

        param_path = os.path.join(_forcefield_directory_path, "protein-allhdg5-4_new.param")
        with open(param_path, encoding="utf-8") as f:
            self._vanderwaals_parameters = ParamParser.parse(f)

    def _find_matching_residue_class(self, residue: Residue) -> str | None:
        for criterium in self._residue_class_criteria:
            if criterium.matches(
                residue.amino_acid.three_letter_code,
                [atom.name for atom in residue.atoms],
            ):
                return criterium.class_name

        return None

    def get_vanderwaals_parameters(self, atom: Atom) -> VanderwaalsParam:
        atom_name = atom.name

        if atom.residue.amino_acid is None:
            _log.warning(f"no amino acid for {atom}; three letter code set to XXX")
            residue_name = "XXX"
        else:
            residue_name = atom.residue.amino_acid.three_letter_code

        type_ = None

        # check top
        top_key = (residue_name, atom_name)
        if top_key in self._top_rows:
            type_ = self._top_rows[top_key]["type"]

        # check patch, which overrides top
        residue_class = self._find_matching_residue_class(atom.residue)
        if residue_class is not None:
            for action in self._patch_actions:
                if action.type in [PatchActionType.MODIFY, PatchActionType.ADD] and residue_class == action.selection.residue_type and "TYPE" in action:
                    type_ = action["TYPE"]

        if type_ is None:
            _log.warning(
                f"Atom {atom} is unknown to the forcefield, vanderwaals_parameters set to (0.0, 0.0, 0.0, 0.0).\n"
                "   This will affect `vanderwaals` feature.\n"
                "       Check https://deeprank2.readthedocs.io/en/latest/features.html#nonbond-energies for more details.",
            )
            return VanderwaalsParam(0.0, 0.0, 0.0, 0.0)
        return self._vanderwaals_parameters[type_]

    def get_charge(self, atom: Atom) -> float:
        """Get the charge of a given `Atom`.

        Args:
            atom(Atom): the atom to get the charge for
        Returns(float): the charge of the given atom.
        """
        atom_name = atom.name
        amino_acid_code = atom.residue.amino_acid.three_letter_code

        charge = None

        # check top
        top_key = (amino_acid_code, atom_name)
        if top_key in self._top_rows:
            charge = float(self._top_rows[top_key]["charge"])

        # check patch, which overrides top
        residue_class = self._find_matching_residue_class(atom.residue)
        if residue_class is not None:
            for action in self._patch_actions:
                if action.type in [PatchActionType.MODIFY, PatchActionType.ADD] and residue_class == action.selection.residue_type:
                    charge = float(action["CHARGE"])

        if charge is None:
            _log.warning(
                f"Atom {atom} is unknown to the forcefield, `electrostatic` and `atom_charge` charge is set to 0.0.\n"
                "   This will affect `electrostatic` and `atom_charge` features.\n"
                "       Check https://deeprank2.readthedocs.io/en/latest/features.html#nonbond-energies for more details.",
            )
            return 0.0
        return charge


atomic_forcefield = AtomicForcefield()



from enum import Enum

import numpy as np
from numpy.typing import NDArray
from typing_extensions import Self


class Polarity(Enum):
    """One-hot encoding of the amino acid polarity."""

    NONPOLAR = 0
    POLAR = 1
    NEGATIVE = 2
    POSITIVE = 3

    @property
    def onehot(self) -> NDArray:
        t = np.zeros(4)
        t[self.value] = 1.0
        return t


class AminoAcid:
    """An amino acid represents the type of `Residue` in a `PDBStructure`.

    Args:
        name (str): Full name of the amino acid.
        three_letter_code (str): Three-letter code of the amino acid (as in PDB).
        one_letter_code (str): One-letter of the amino acid (as in fasta).
        charge (int): Charge of the amino acid.
        polarity (:class:`Polarity`): The polarity of the amino acid.
        size (int): The number of non-hydrogen atoms in the side chain.
        mass (float): Average residue mass (i.e. mass of amino acid - H20) in Daltons.
        pI (float): Isolectric point; pH at which the molecule has no net electric charge.
        hydrogen_bond_donors (int): Number of hydrogen bond donors.
        hydrogen_bond_acceptors (int): Number of hydrogen bond acceptors.
        index (int): The rank of the amino acid, used for computing one-hot encoding.
    """

    def __init__(
        self,
        name: str,
        three_letter_code: str,
        one_letter_code: str,
        charge: int,
        polarity: Polarity,
        size: int,
        mass: float,
        pI: float,
        hydrogen_bond_donors: int,
        hydrogen_bond_acceptors: int,
        index: int,
    ):
        # amino acid nomenclature
        self._name = name
        self._three_letter_code = three_letter_code
        self._one_letter_code = one_letter_code

        # side chain properties
        self._charge = charge
        self._polarity = polarity
        self._size = size
        self._mass = mass
        self._pI = pI
        self._hydrogen_bond_donors = hydrogen_bond_donors
        self._hydrogen_bond_acceptors = hydrogen_bond_acceptors

        # one hot encoding
        self._index = index

    @property
    def name(self) -> str:
        return self._name

    @property
    def three_letter_code(self) -> str:
        return self._three_letter_code

    @property
    def one_letter_code(self) -> str:
        return self._one_letter_code

    @property
    def charge(self) -> int:
        return self._charge

    @property
    def polarity(self) -> Polarity:
        return self._polarity

    @property
    def size(self) -> int:
        return self._size

    @property
    def mass(self) -> float:
        return self._mass

    @property
    def pI(self) -> float:
        return self._pI

    @property
    def hydrogen_bond_donors(self) -> int:
        return self._hydrogen_bond_donors

    @property
    def hydrogen_bond_acceptors(self) -> int:
        return self._hydrogen_bond_acceptors

    @property
    def onehot(self) -> NDArray:
        if self._index is None:
            msg = f"Amino acid {self._name} index is not set, thus no onehot can be computed."
            raise ValueError(msg)
        # 20 canonical amino acids
        # selenocysteine and pyrrolysine are indexed as cysteine and lysine, respectively
        a = np.zeros(20)
        a[self._index] = 1.0
        return a

    @property
    def index(self) -> int:
        return self._index

    def __hash__(self) -> hash:
        return hash(self.name)

    def __eq__(self, other: Self) -> bool:
        if isinstance(other, AminoAcid):
            return other.name == self.name
        return NotImplemented

    def __repr__(self) -> str:
        return self._three_letter_code



import glob
import logging
import os

import h5py
import numpy as np
from pdb2sql import StructureSimilarity

from deeprank2.domain import targetstorage as targets

_log = logging.getLogger(__name__)
MIN_IRMS_FOR_BINARY = 4


def add_target(  # noqa: C901
    graph_path: str | list[str],
    target_name: str,
    target_list: str,
    sep: str = " ",
) -> None:
    """Add a target to all the graphs in hdf5 files.

    Args:
        graph_path (str | list(str)): Either a directory containing all the hdf5 files,
            or a single hdf5 filename
            or a list of hdf5 filenames.
        target_name (str): The name of the new target.
        target_list (str): Name of the file containing the data.
        sep (str, optional): Separator in target list. Defaults to " ".

    Notes:
        The input target list should respect the following format :
        1ATN_xxx-1 0
        1ATN_xxx-2 1
        1ATN_xxx-3 0
        1ATN_xxx-4 0
    """
    target_dict = {}

    labels = np.loadtxt(target_list, delimiter=sep, usecols=[0], dtype=str)
    values = np.loadtxt(target_list, delimiter=sep, usecols=[1])
    for label, value in zip(labels, values, strict=True):
        target_dict[label] = value

    if os.path.isdir(graph_path):
        graphs = glob.glob(f"{graph_path}/*.hdf5")
    elif os.path.isfile(graph_path):
        graphs = [graph_path]
    elif isinstance(graph_path, list):
        graphs = graph_path
    else:
        msg = "Incorrect input passed."
        raise TypeError(msg)

    for hdf5 in graphs:
        _log.info(hdf5)
        if not os.path.isfile(hdf5):
            msg = f"File {hdf5} not found."
            raise FileNotFoundError(msg)

        try:
            f5 = h5py.File(hdf5, "a")
            for model in target_dict:
                if model not in f5:
                    msg = f"{hdf5} does not contain an entry named {model}."
                    raise ValueError(msg)  # noqa: TRY301
                try:
                    model_gp = f5[model]
                    if targets.VALUES not in model_gp:
                        model_gp.create_group(targets.VALUES)
                    group = f5[f"{model}/{targets.VALUES}/"]
                    if target_name in group:
                        # Delete the target if it already existed
                        del group[target_name]
                    # Create the target
                    group.create_dataset(target_name, data=target_dict[model])
                except BaseException:  # noqa: BLE001
                    _log.info(f"no graph for {model}")
            f5.close()

        except BaseException:  # noqa: BLE001
            _log.info(f"no graph for {hdf5}")


def compute_ppi_scores(
    pdb_path: str,
    reference_pdb_path: str,
) -> dict[str, float | int]:
    """Compute structure similarity scores for the input docking model and return them as a dictionary.

    The computed scores are: `lrmsd` (ligand root mean square deviation), `irmsd` (interface rmsd),
    `fnat` (fraction of native contacts), `dockq` (docking model quality), `binary` (True - high quality,
    False - low quality), `capri_class` (capri classification, 1 - high quality, 2 - medium, 3 - acceptable,
    4 - incorrect). See https://deeprank2.readthedocs.io/en/latest/docking.html for more details about the scores.

    Args:
        pdb_path (str): Path to the decoy.
        reference_pdb_path (str): Path to the reference (native) structure.

    Returns: a dictionary containing values for lrmsd, irmsd, fnat, dockq, binary, capri_class.
    """
    ref_name = os.path.splitext(os.path.basename(reference_pdb_path))[0]
    sim = StructureSimilarity(
        pdb_path,
        reference_pdb_path,
        enforce_residue_matching=False,
    )

    scores = {}

    # Input pre-computed zone files
    if os.path.exists(ref_name + ".lzone"):
        scores[targets.LRMSD] = sim.compute_lrmsd_fast(method="svd", lzone=ref_name + ".lzone")
        scores[targets.IRMSD] = sim.compute_irmsd_fast(method="svd", izone=ref_name + ".izone")

    # Compute zone files
    else:
        scores[targets.LRMSD] = sim.compute_lrmsd_fast(method="svd")
        scores[targets.IRMSD] = sim.compute_irmsd_fast(method="svd")

    scores[targets.FNAT] = sim.compute_fnat_fast()
    scores[targets.DOCKQ] = sim.compute_DockQScore(scores[targets.FNAT], scores[targets.LRMSD], scores[targets.IRMSD])
    scores[targets.BINARY] = scores[targets.IRMSD] < MIN_IRMS_FOR_BINARY

    scores[targets.CAPRI] = 4
    for thr, val in zip([4.0, 2.0, 1.0], [3, 2, 1], strict=True):
        if scores[targets.IRMSD] < thr:
            scores[targets.CAPRI] = val

    return scores



from enum import Enum
from pathlib import Path

import numpy as np
from Bio.PDB import PDBParser
from Bio.PDB.DSSP import DSSP
from numpy.typing import NDArray

from deeprank2.domain import nodestorage as Nfeat
from deeprank2.molstruct.atom import Atom
from deeprank2.molstruct.residue import Residue, SingleResidueVariant
from deeprank2.utils.graph import Graph


class DSSPError(Exception):
    """Raised if DSSP fails to produce an output."""


class SecondarySctructure(Enum):
    """Value to express a secondary a residue's secondary structure type."""

    HELIX = 0  # 'GHI'
    STRAND = 1  # 'BE'
    COIL = 2  # ' -STP'

    @property
    def onehot(self) -> NDArray:
        t = np.zeros(3)
        t[self.value] = 1.0

        return t


def _get_records(lines: list[str]) -> list[str]:
    seen = set()
    seen_add = seen.add
    return [x.split()[0] for x in lines if not (x in seen or seen_add(x))]


def _check_pdb(pdb_path: str) -> None:
    """Check whether pdb metadata required for DSSP exists and auto-fix in place where possible.

    Args:
        pdb_path: file location of pdb file
    """
    fix_pdb = False
    with open(pdb_path, encoding="utf-8") as f:
        lines = f.readlines()

    # check for HEADER
    firstline = lines[0]
    if not firstline.startswith("HEADER"):
        fix_pdb = True
        if firstline.startswith("EXPDTA"):
            lines = [f"HEADER {firstline}"] + lines[1:]
        else:
            lines = ["HEADER \n", *lines]

    # check for CRYST1 record
    existing_records = _get_records(lines)
    if "CRYST1" not in existing_records:
        fix_pdb = True
        dummy_CRYST1 = "CRYST1   00.000   00.000   00.000  00.00  00.00  00.00 X 00 00 0    00\n"
        lines = [lines[0]] + [dummy_CRYST1] + lines[1:]

    # check for unnumbered REMARK lines
    for i, line in enumerate(lines):
        if line.startswith("REMARK"):
            try:
                int(line.split()[1])
            except ValueError:
                fix_pdb = True
                lines[i] = f"REMARK 999 {line[7:]}"

    if fix_pdb:
        with open(pdb_path, "w", encoding="utf-8") as f:
            f.writelines(lines)


def _classify_secstructure(subtype: str) -> SecondarySctructure:
    if subtype in "GHI":
        return SecondarySctructure.HELIX
    if subtype in "BE":
        return SecondarySctructure.STRAND
    if subtype in " -STP":
        return SecondarySctructure.COIL
    return None


def _get_secstructure(pdb_path: str) -> dict:
    """Process the DSSP output to extract secondary structure information.

    Args:
        pdb_path (str): The file path of the PDB file to be processed.

    Returns:
        dict: A dictionary containing secondary structure information for each chain and residue.
    """
    # Execute DSSP and read the output
    _check_pdb(pdb_path)
    p = PDBParser(QUIET=True)
    model = p.get_structure(Path(pdb_path).stem, pdb_path)[0]

    try:
        dssp = DSSP(model, pdb_path, dssp="mkdssp")
    except Exception as e:
        pdb_format_link = "https://www.wwpdb.org/documentation/file-format-content/format33/sect1.html#Order"
        msg = (
            f"DSSP has raised the following exception: {e}.\n\t"
            f"This is likely due to an improrperly formatted pdb file: {pdb_path}.\n\t"
            f"See {pdb_format_link} for guidance on how to format your pdb files.\n\t"
            "Alternatively, turn off secondary_structure feature module during QueryCollection.process()."
        )
        raise DSSPError(msg) from e

    chain_ids = [dssp_key[0] for dssp_key in dssp.property_keys]
    res_numbers = [dssp_key[1][1] for dssp_key in dssp.property_keys]
    sec_structs = [dssp[dssp_key][2] for dssp_key in dssp.property_keys]

    # Store output in Dictionary
    sec_structure_dict = {}
    for chain in set(chain_ids):
        sec_structure_dict[chain] = {}
    for i, _ in enumerate(chain_ids):
        sec_structure_dict[chain_ids[i]][res_numbers[i]] = sec_structs[i]

    return sec_structure_dict


def add_features(  # noqa:D103
    pdb_path: str,
    graph: Graph,
    single_amino_acid_variant: SingleResidueVariant | None = None,  # noqa: ARG001
) -> None:
    sec_structure_features = _get_secstructure(pdb_path)

    for node in graph.nodes:
        if isinstance(node.id, Residue):
            residue = node.id
        elif isinstance(node.id, Atom):
            atom = node.id
            residue = atom.residue
        else:
            msg = f"Unexpected node type: {type(node.id)}"
            raise TypeError(msg)

        chain_id = residue.chain.id
        res_num = residue.number

        try:
            node.features[Nfeat.SECSTRUCT] = _classify_secstructure(sec_structure_features[chain_id][res_num]).onehot
        except AttributeError as e:
            msg = f"Unknown secondary structure type ({sec_structure_features[chain_id][res_num]}) detected on chain {chain_id} residues {res_num}."
            raise ValueError(msg) from e



class VanderwaalsParam:  # noqa: D101
    def __init__(
        self,
        epsilon_main: float,
        sigma_main: float,
        epsilon_14: float,
        sigma_14: float,
    ):
        self.epsilon_main = epsilon_main
        self.sigma_main = sigma_main
        self.epsilon_14 = epsilon_14
        self.sigma_14 = sigma_14

    def __str__(self) -> str:
        return f"{self.epsilon_main}, {self.sigma_main}, {self.epsilon_14}, {self.sigma_14}"


class ParamParser:  # noqa: D101
    @staticmethod
    def parse(file_: str) -> dict[str, VanderwaalsParam]:
        result = {}
        for line in file_:
            if line.startswith("#"):
                continue

            if line.startswith("NONBonded "):
                (
                    _,
                    type_,
                    epsilon_main,
                    sigma_main,
                    epsilon_14,
                    sigma_14,
                ) = line.split()

                result[type_] = VanderwaalsParam(
                    float(epsilon_main),
                    float(sigma_main),
                    float(epsilon_14),
                    float(sigma_14),
                )
            elif len(line.strip()) == 0:
                continue
            else:
                msg = f"Unparsable param line: {line}"
                raise ValueError(msg)

        return result



from typing_extensions import Self

from deeprank2.molstruct.aminoacid import AminoAcid
from deeprank2.molstruct.residue import Residue


class PssmRow:
    """Holds data for one position-specific scoring matrix row."""

    def __init__(
        self,
        conservations: dict[AminoAcid, float],
        information_content: float,
    ):
        self._conservations = conservations
        self._information_content = information_content

    @property
    def conservations(self) -> dict[AminoAcid, float]:
        return self._conservations

    @property
    def information_content(self) -> float:
        return self._information_content

    def get_conservation(self, amino_acid: AminoAcid) -> float:
        return self._conservations[amino_acid]


class PssmTable:
    """Holds data for one position-specific scoring table."""

    def __init__(self, rows: list[PssmRow] | None = None):
        if rows is None:
            self._rows = {}
        else:
            self._rows = rows

    def __contains__(self, residue: Residue) -> bool:
        return residue in self._rows

    def __getitem__(self, residue: Residue) -> PssmRow:
        return self._rows[residue]

    def update(self, other: Self) -> None:
        """Can be used to merge two non-overlapping scoring tables."""
        self._rows.update(other._rows)  # noqa: SLF001



import torch
import torch.nn.functional as F
from torch import nn
from torch_geometric.nn import max_pool_x
from torch_geometric.nn.inits import uniform
from torch_scatter import scatter_mean, scatter_sum

from deeprank2.utils.community_pooling import community_pooling, get_preloaded_cluster

# ruff: noqa: ANN001, ANN201


class GINetConvLayer(torch.nn.Module):  # noqa: D101
    def __init__(self, in_channels, out_channels, number_edge_features=1, bias=False):
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels

        self.fc = nn.Linear(self.in_channels, self.out_channels, bias=bias)
        self.fc_edge_attr = nn.Linear(number_edge_features, number_edge_features, bias=bias)
        self.fc_attention = nn.Linear(2 * self.out_channels + number_edge_features, 1, bias=bias)
        self.reset_parameters()

    def reset_parameters(self) -> None:
        size = self.in_channels
        uniform(size, self.fc.weight)
        uniform(size, self.fc_attention.weight)
        uniform(size, self.fc_edge_attr.weight)

    def forward(self, x, edge_index, edge_attr):
        row, col = edge_index
        num_node = len(x)
        edge_attr = edge_attr.unsqueeze(-1) if edge_attr.dim() == 1 else edge_attr

        xcol = self.fc(x[col])
        xrow = self.fc(x[row])

        ed = self.fc_edge_attr(edge_attr)
        # create edge feature by concatenating node feature
        alpha = torch.cat([xrow, xcol, ed], dim=1)
        alpha = self.fc_attention(alpha)
        alpha = F.leaky_relu(alpha)

        alpha = F.softmax(alpha, dim=1)
        h = alpha * xcol

        out = torch.zeros(num_node, self.out_channels).to(alpha.device)
        z = scatter_sum(h, row, dim=0, out=out)

        return z  # noqa:RET504 (unnecessary-assign)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.in_channels}, {self.out_channels})"


class GINet(torch.nn.Module):  # noqa: D101
    # input_shape -> number of node input features
    # output_shape -> number of output value per graph
    # input_shape_edge -> number of edge input features
    def __init__(self, input_shape, output_shape=1, input_shape_edge=1):
        super().__init__()
        self.conv1 = GINetConvLayer(input_shape, 16, input_shape_edge)
        self.conv2 = GINetConvLayer(16, 32, input_shape_edge)

        self.conv1_ext = GINetConvLayer(input_shape, 16, input_shape_edge)
        self.conv2_ext = GINetConvLayer(16, 32, input_shape_edge)

        self.fc1 = nn.Linear(2 * 32, 128)
        self.fc2 = nn.Linear(128, output_shape)
        self.clustering = "mcl"
        self.dropout = 0.4

    def forward(self, data):
        act = F.relu
        data_ext = data.clone()

        # EXTERNAL INTERACTION GRAPH
        # first conv block
        data.x = act(self.conv1(data.x, data.edge_index, data.edge_attr))
        cluster = get_preloaded_cluster(data.cluster0, data.batch)
        data = community_pooling(cluster, data)

        # second conv block
        data.x = act(self.conv2(data.x, data.edge_index, data.edge_attr))
        cluster = get_preloaded_cluster(data.cluster1, data.batch)
        x, batch = max_pool_x(cluster, data.x, data.batch)

        # INTERNAL INTERACTION GRAPH
        # first conv block
        data_ext.x = act(self.conv1_ext(data_ext.x, data_ext.edge_index, data_ext.edge_attr))
        cluster = get_preloaded_cluster(data_ext.cluster0, data_ext.batch)
        data_ext = community_pooling(cluster, data_ext)

        # second conv block
        data_ext.x = act(self.conv2_ext(data_ext.x, data_ext.edge_index, data_ext.edge_attr))
        cluster = get_preloaded_cluster(data_ext.cluster1, data_ext.batch)
        x_ext, batch_ext = max_pool_x(cluster, data_ext.x, data_ext.batch)

        # FC
        x = scatter_mean(x, batch, dim=0)
        x_ext = scatter_mean(x_ext, batch_ext, dim=0)

        x = torch.cat([x, x_ext], dim=1)
        x = act(self.fc1(x))
        x = F.dropout(x, self.dropout, training=self.training)
        x = self.fc2(x)

        return x  # noqa:RET504 (unnecessary-assign)



import logging
import re
from typing import Any

logging.getLogger(__name__)


class TopRowObject:  # noqa: D101
    def __init__(
        self,
        residue_name: str,
        atom_name: str,
        kwargs: dict[str, Any],
    ):
        self.residue_name = residue_name
        self.atom_name = atom_name
        self.kwargs = kwargs

    def __getitem__(self, key: str):
        return self.kwargs[key]


class TopParser:  # noqa: D101
    _VAR_PATTERN = re.compile(r"([^\s]+)\s*=\s*([^\s\(\)]+|\(.*\))")
    _LINE_PATTERN = re.compile(r"^([A-Z0-9]{3})\s+atom\s+([A-Z0-9]{1,4})\s+(.+)\s+end\s*(\s+\!\s+[ _A-Za-z0-9]+)?$")
    _NUMBER_PATTERN = re.compile(r"\-?[0-9]+(\.[0-9]+)?")

    @staticmethod
    def parse(file_: str) -> list[TopRowObject]:
        result = []
        for line in file_:
            # parse the line
            m = TopParser._LINE_PATTERN.match(line)
            if not m:
                msg = f"Unmatched top line: {line}"
                raise ValueError(msg)

            residue_name = m.group(1).upper()
            atom_name = m.group(2).upper()

            kwargs = {}
            for w in TopParser._VAR_PATTERN.finditer(m.group(3)):
                kwargs[w.group(1).lower().strip()] = TopParser._parse_value(w.group(2).strip())

            result.append(TopRowObject(residue_name, atom_name, kwargs))

        return result

    @staticmethod
    def _parse_value(s: str) -> float | str:
        # remove parentheses
        if s[0] == "(" and s[-1] == ")":
            return TopParser._parse_value(s[1:-1])

        if TopParser._NUMBER_PATTERN.match(s):
            return float(s)

        return s



import logging
import signal
import sys
import warnings
from typing import NoReturn

import numpy as np
from Bio.PDB.Atom import PDBConstructionWarning
from Bio.PDB.HSExposure import HSExposureCA
from Bio.PDB.PDBParser import PDBParser
from Bio.PDB.ResidueDepth import get_surface, residue_depth

from deeprank2.domain import nodestorage as Nfeat
from deeprank2.molstruct.atom import Atom
from deeprank2.molstruct.residue import Residue, SingleResidueVariant
from deeprank2.utils.graph import Graph

_log = logging.getLogger(__name__)


def handle_sigint(sig, frame) -> None:  # noqa: ARG001, ANN001, D103
    _log.info("SIGINT received, terminating.")
    sys.exit()


def handle_timeout(sig, frame) -> NoReturn:  # noqa: ARG001, ANN001, D103
    msg = "Timed out!"
    raise TimeoutError(msg)


def space_if_none(value: str) -> str:  # noqa:D103
    if value is None:
        return " "
    return value


def add_features(  # noqa:D103
    pdb_path: str,
    graph: Graph,
    single_amino_acid_variant: SingleResidueVariant | None = None,  # noqa: ARG001
) -> None:
    signal.signal(signal.SIGINT, handle_sigint)
    signal.signal(signal.SIGALRM, handle_timeout)

    with warnings.catch_warnings(record=PDBConstructionWarning):
        parser = PDBParser()
        structure = parser.get_structure("_tmp", pdb_path)
    bio_model = structure[0]

    try:
        signal.alarm(20)
        surface = get_surface(bio_model)
        signal.alarm(0)
    except TimeoutError as e:
        msg = "Bio.PDB.ResidueDepth.get_surface timed out."
        raise TimeoutError(msg) from e

    # These can only be calculated per residue, not per atom.
    # So for atomic graphs, every atom gets its residue's value.
    hse = HSExposureCA(bio_model)
    for node in graph.nodes:
        if isinstance(node.id, Residue):
            residue = node.id
        elif isinstance(node.id, Atom):
            atom = node.id
            residue = atom.residue
        else:
            msg = f"Unexpected node type: {type(node.id)}"
            raise TypeError(msg)

        bio_residue = bio_model[residue.chain.id][residue.number]
        node.features[Nfeat.RESDEPTH] = residue_depth(bio_residue, surface)
        hse_key = (
            residue.chain.id,
            (" ", residue.number, space_if_none(residue.insertion_code)),
        )

        if hse_key in hse:
            node.features[Nfeat.HSE] = np.array(hse[hse_key], dtype=np.float64)
        else:
            node.features[Nfeat.HSE] = np.array((0, 0, 0), dtype=np.float64)
